/*
	-- DesktopImage.cc		V2.0

	------------------------------------------------------------------------------
	-- Description:

		This custom component is used to display an image on VdB's desktop area
      (or in any deployed application's MDI desktop). 
      The "desktop area" is the grey zone which stands between the top (toolbar 
      or menu bar) and the bottom (or the status bar) of VdB's IDE main window,
      or your application parent MDI window.
		As a subclass of VdB's image class, its datasourced image will be displayed 
		on the desktop. Several properties and methods let you control the image's
		position on the desktop, while the standard image class properties let you
		control the appearance.
		All types of images supported by VdB's image class can be displayed.

	-- Programmer:

			Nicolas Martin,          nicolas.martin@freesurf.fr

	-- Copyright Nicolas Martin, 1999

	-- History:

			07/09/1999 -- Original V1.0
			10/09/1999 -- V2.0
         	- Name changed
				- Change superclass to image class
				- Added a getDesktopSize method and two properties to surface the desktop
				  area size
				- Modified the moveImage method to avoid flickering
				- Added 2 optional parameters nLeft and nTop to the showImage method so
				  that the image can be sticked anywhere in the desktop area.
				- Major comments and documentation modifications

	-- designed for Visual dBase, version 7.01,
		tested under Win 95, in VdB's IDE and MDI application.

	-- dependancies:

				- Structure.prg
				- StructAPI.h
				- StructAPI.prg
				- structmem.dll
				- Winuser.h

	-- How to use:
		-----------

		You can add this component to your component palette.
		The image to display must be contained in a VdB's image object.
		Add the DesktopImage component to the form or container that holds
		this image object that will be displayed on the desktop. The easiest 
		way to do that is under the form designer, as you can directly adjust 
		the appearance (size, border type...) of the image to display (See the 
		file DemoSetImage.wfm).
		Then, use the showImage method to display the image, in the form's open()
		method.

		NOTE: All image positions are referenced in pixels; be sure to set the
		metric property to "pixels" in the parent form.

		Programmatically, this can be done like this :

			set procedure to DesktopImage.cc additive
			local f
			f = new form()
			f.visible = false
			f.metric = 6 	// pixels
			oDImg = new DesktopImage(f)
			oDImg.datasource = "filename " + _dbwinhome + "artwork\splash\factory.bmp"
			oDImg.alignment = 4	// True Size
			f.open()
			oDImg.showImage( )

	-- Properties:
		----------
		As this component is itself subclassed from VdB's standard image class,
		all the properties that concern the image appearance should be used
		as for standard images display with forms. Consult the image class in the
		online help.
		The following properties have been added for the specific use to display 
		on the MDI desktop:

		.ImageLeft: VdB numeric, the position of the image's left side, in 
						pixels, referenced to the MDI parent client area (the grey
						desktop). This property is read only.

		.ImageTop:  VdB numeric, the position of the image's top side, in 
						pixels, referenced to the MDI parent client area (the grey
						desktop). This property is read only.

		.DesktopWidth: VdB numeric, gives the width of the desktop area as returned
							by the last call to the getDesktopSize() method.
							This property is read only.

		.DesktopHeight: VdB numeric, gives the height of the desktop area as returned
							by the last call to the getDesktopSize() method.
							This property is read only.


	-- Methods:
		-------
		getDesktopSize()
			Updates the 2 properties .DesktopWidth and .DesktopHeight with the current
			desktop area size. This method is to be called whenever the parent MDI 
			window is resized, modifying consequently the desktop area size (and if this
			new size is needed in your code).

		showImage( [<nLeft>, <nTop>] )
			This method displays the image on the MDI desktop, centered, if no 
			nLeft and nTop coordinates are specified.
			If nLeft and nTop are specified, the image top left corner will be positioned 
			at these coordinates (in pixels), relatively to the desktop area top left
			corner.
			The 2 properties .ImageLeft and .ImageTop are updated.

			<nLeft>,<nTop>: Optional, VdB numeric. If not specified, the image is 
								 centered.
								 If specified, the top left corner of the image is set 
								 up to these coordinates, expressed in pixels, relatively 
								 from the top left corner of the desktop area.

			return : nothing

		hideImage( <oImage> )
			This method hides the image referenced by oImage

			<oImage>: VdB image object reference, that contains the image reference.

			return : nothing

		moveImage( <nLeft>, <nTop> )
			This method moves the image to the new location on the desktop specified 
			by <nLeft> and <nTop>. The 2 properties .ImageLeft and .ImageTop are updated.

			<nLeft>:  VdB numeric, the new left side position of the image, in pixels

			<nTop>:   VdB numeric, the new top side position of the image, in pixels

			return : nothing

	-- Run in VdB IDE and Deployment information:
		-----------------------------------------

		1/ You need to have the samples directory installed as part of your
			VdB 7.01 installation.
		2/ To build an executable that includes DesktopImage, you need to:
			- add the following files to your project .prj description file:
			   DesktopImage.cc
			   structure.prg      (Both are in the samples directory; leave them there
			   structAPI.prg       simply reference them in the .prj file)
			- place a copy of the following .dll in the executable directory
				(you need to add it to your deployed application files)
			   structmem.dll	     (it is in the samples directory)


	------------------------------------------------------------------------------
*/

#include <Winuser.h>
#include <StructAPI.h>

// The DesktopImage class is inherited from the Image class
class DesktopImage( f, n ) of Image( f, n ) custom
	protect nMDIClientHwnd, lpRect

	// used for structure API stuff
	set procedure to '&_dbwinhome.samples\structure.prg' additive

	// Reference some externs
	extern CHANDLE        SetParent(CHANDLE, CHANDLE) user32
	extern CINT           SetWindowPos(CHANDLE, CHANDLE, CINT, CINT, CINT, ;
								 CINT, UINT) user32
	extern CHANDLE        GetWindow(CHANDLE, CINT) user32
	extern BOOL           EnableWindow(CHANDLE, BOOL) user32
	extern BOOL           GetClientRect(CHWND, LPRECT) user32
	extern HANDLE         FindWindowEx(HANDLE, HANDLE, LPCSTR, LPCSTR) user32 ;
								 from "FindWindowExA"
	extern BOOL        	 MoveWindow(HANDLE, CINT, CINT, CINT, CINT, BOOL) user32


	// The following line of code is the key point of the component: 
	// Like other MDI applications, VdB 7.01 uses a parent window, whose Win32 
	// classname is set to <BMain>, and handle is _app.framewin.hwnd.
	// This parent window contains a child window, Win32 classname is <MDIClient>,
	// that graphically represents the desktop area. In order to display other
	// stuff in that window, we need to retreive the handle of that window.
	// The images (and probably any other widgets) can then be attached as children 
	// to that window.
	// So first we retreive the handle to this <MDIClient> window:
	this.nMDIClientHwnd = FindWindowEx( _app.Framewin.hwnd, Null, "MDIClient", Null )

	// The following structure will be used to retreive the current size of the
	// desktop area:
	this.lpRect = new _RECT()      // create a _RECT structure
	this.DesktopWidth = 0
	this.DesktopHeight = 0
	this.getDesktopSize( )			 // updates .DesktopWidth and .DesktopHeight properties

	// Initialize the image position relative to desktop properties.
	this.ImageLeft = 0
	this.ImageTop = 0

	// Set the visible property to false to avoid flickering when opening
	this.visible := false

	// The following method retreives the size of the desktop area and updates the 
	//	.DesktopWidth and .DesktopHeight properties
	function getDesktopSize
		// Retreives the dimensions of the desktop client area
		GetClientRect( this.nMDIClientHwnd, this.lpRect.value )
		// update the properties
		this.DesktopWidth = this.lpRect.getMember( "right" ) - this.lpRect.getMember( "left" )
		this.DesktopHeight = this.lpRect.getMember( "bottom" ) - this.lpRect.getMember( "top" )
		return

	// This method displays the image. Centered if no parameter is passed,
	// top left corner set to (nLeft, nTop) if those 2 parameters are supplied.
	function showImage( nLeft, nTop )
		if argcount( ) == 0			// image will be centered
			this.getDesktopSize( )
			// Centered : compute the coordinates of the top left corner where the image
			// is to be attached 
			this.ImageLeft = ( this.DesktopWidth - this.width ) / 2
			this.ImageTop = ( this.DesktopHeight - this.height ) / 2
		else 								// Coordinates specified
			this.ImageLeft = nLeft
			this.ImageTop = nTop
		endif

		// Attach the image to the desktop area, using the image handle and the 
		// desktop (MDIClient) handle previously found.
		SetParent( this.hwnd, this.nMDIClientHwnd )

		// This will disables the image window so that it will be unsensitive to 
		// any mouse & keyboard events. This will especially prevent the image from 
		// getting focus.
		EnableWindow( this.hwnd, false )

		// Sets the position of the image, at bottom Z order so that it will
		// appear beneath any other child window of the parent window
		SetWindowPos( this.hwnd, ;
							HWND_BOTTOM, ;
							this.ImageLeft, ;
							this.ImageTop, ;
							this.width, ;
							this.height, ;
							0 )

		// Displays the image, set its visible property to true
		this.visible := true
		return

	// This trivial method will hide the image. This shows that even if the image's
	// parent is now the desktop area, its properies are still accessible as usual.
	function hideImage( oImage )
		this.visible := false
		return

	// This method moves the image to a new position on the desktop area
	function moveImage( nLeft, nTop )
		// Updates the image position properties
		this.ImageLeft := nLeft
		this.ImageTop := nTop

		// moves the image, and repaint the parent window
		MoveWindow( this.hwnd, nLeft, nTop, this.width, this.height, true )
   	return

endclass

// This class defines the API _RECT structure, used here in the GetClientRect call
class _RECT of structure    // rc  
    super::addMember( TYPE_DWORD, "left" ) 
    super::addMember( TYPE_DWORD, "top" )
    super::addMember( TYPE_DWORD, "right" ) 
	 super::addMember( TYPE_DWORD, "bottom" ) 
endclass
