// Metric: 0 - Chars only
/*
   FLASHBTN.CC

   // NOTE: This file includes both FLASHBTN and ZFORM (a custom
   // form class with special properties that can be used
   // by the FLASHBTN ...

  Reverses the color of the pushbutton when the mouse
  cursor passes over it.
 
  Author: Gary White  garyw@qni.com
  
  Last modified: Oct 06, 1997
  Then by Peter Rorlick on March 2, 1998
 
  Usage: In the form designer, click File, Set up custom controls
  and add the FlashButton.CC.  Then just select the FlashButton
  from the Custom tab of your control palette and install it just
  like you would any other button.
 
  This custom class uses the pushbutton's OnOpen event to install
  the routines.  Do NOT overwrite the FlashButton's OnOpen event.
  The OnMouseMove event triggers the color change and is protected.
 
  Note that any controls created after the form's OnOpen will not
  have the routine installed to return the FlashButton to it's
  normal colors.  This should rarely be a problem unless the newly
  created control is touching the FlashButton, in which case the
  button will not return to normal until the mouse pointer moves
  off of the new control.  Unlikely, but it could happen.
 
  Thanks to Romain Strieff and Peter Rorlick for finding my errors
  and suggesting fixes and improvements.
 
  Enjoy!
  Gary
 
  -----
  Changes made by Peter Rorlick on March 2nd:

  We now test whether a control is a Flashbutton or one of its
  subclasses:
     if Type('oControl.HotColor')=='U'
  This allows us to subclass it without it breaking.

  I've modified the Do...Until loop so that it recurses through
  the objects that belong to containers and notebooks.
  This solves the small glitch that occurred when a
  flashbutton resided on a container or notebook and there was
  another control touching it: when the mouse moved from the
  flashbutton to the other control, the button did not un-flash.

  The March 2 version now sets a specific hot color (Black/Yellow) 
  and changes the borderstyle, instead of simply reversing the 
  color. I find that it looks much nicer. But if you have small 
  buttons with bitmaps, you might prefer to comment out the lines 
  that change the borderstyle (or set borderStyle to 4 rather than 8).

  We now use an Init method instead of OnOpen, so that we don't have
  to worry about programmers over-writing the OnOpen.  To take 
  advantage of the Init method, the form must be based on zForm or 
  one of its subclasses. If not based on zForm, it uses the 
  OnOpen method.

  The button also lights up now when the user tabs to it.

*/
CLASS FLASHBUTTON( FormObj ) OF PUSHBUTTON( FormObj ) Custom

   protect OnOpen, OnMouseMove
   this.Height      := 1.3
   this.Width       := 12
   this.Text        := "FlashBtn"
   this.FontBold    := false
   this.OnMouseMove := class::FlashButton_OnMouseMove
   this.OnGotFocus  := class::FlashButton_OnMouseMove

   If Type('form.Init') # 'U'
     /* 
        If the form is subclassed from zForm, it will automatically
        run any Init methods of the controls on the form, before the 
        form opens.  In this case, We use the Init method instead of 
        the OnOpen, to "free up" the button's OnOpen for other 
        purposes that may be desired by the programmer.
     */
   Else
     /* 
        Form is not subclassed from zForm. In this case, remember 
        to call Flashbutton::Init() when you over-write the 
        button's OnOpen.
     */
     this.OnOpen = class::Init
   Endif
   this.parent.OnMouseMove = {;class::FlashButton_BlinkOff()}  
     // For Containers and Notebooks that contain flashbuttons.
   form.OnMouseMove = {;class::FlashButton_BlinkOff()}   
    // Needed when the only flashbuttons on the form are on 
    // Containers or Notebooks.


   Procedure FlashButton_OnMouseMove( flags, row, col )
      // Make the button stand out visually:
      if type("form.HotButton") # "O"
        this.ColorNormal := this.HotColor
        this.borderStyle := 8   //    4 is Single, 8 is modal
        form.HotButton    = this
      elseif this.Name # form.HotButton.Name
        form.HotButton.BorderStyle := 0 // back to default
        form.HotButton.ColorNormal := form.HotButton.NormalColor
        this.borderStyle           := 8
        this.ColorNormal           := this.HotColor
        form.HotButton              = this
      endif


   Procedure FlashButton_BlinkOff
      // Return the button's appearance to normal (un-flash it):
      if type("form.HotButton") == "O" and form.HotButton#this
         form.HotButton.ColorNormal := form.HotButton.NormalColor
         form.HotButton.BorderStyle := 0 // back to default
         form.HotButton = ""
      endif

/*
  The Init (OnOpen) routine installs the BlinkOff routine in all other
  controls on the form.  It also determines the NormalColor from
  the ColorNormal assigned by the programmer.
*/

   Procedure Init  // was: FlashButton_OnOpen
      private oControl
      this.NormalColor = this.ColorNormal
      this.HotColor = 'Black/Yellow'
      if type("form.FlashBtnInstalled") == "U"
         form.FlashBtnInstalled = true
         CLASS::Install_BlinkOff(Form)
      endif


   Procedure Install_BlinkOff(InitialObject)
     private o
     o = InitialObject.First
     do
        if ("CONTAINER" $ o.CLASSNAME) OR ("NOTEBOOK" $ o.CLASSNAME)
           CLASS::Install_BlinkOff(o)   
           // recurse through the objects of the container or notebook
        EndIf
        if type( "o.onMouseMove" ) # "U"
           if Type('o.HotColor') == 'U'   
             // If this is NOT a flashbutton or one of its subclasses...
          if empty(o.OnMousemove)
             o.OnMouseMove = {;class::FlashButton_BlinkOff()}
          else
             o.sOnMouseMove = o.OnMouseMove
             o.FlashButton_Blinkoff = {;class::FlashButton_Blinkoff(); this.sOnMouseMove()}
             o.OnMouseMove = {;this.FlashButton_Blinkoff()}
          endif
        endif
      endif
      o = o.before
      until o.name == InitialObject.First.name
   return

ENDCLASS

/*
   ------------------------------------------------------------------
   zForm (base form) class
   Author: Peter Rorlick
   Feb 1998
   Thanks to Romain and Bowen for their ideas.
   ------------------------------------------------------------------
*/
class zForm of FORM custom

   with (this)
      height := 10
      width  := 50
      text   := 'zForm'
      AllowMultipleConcurrentInstances = false   
       // Programmers can set this to TRUE if they like, in their form's 
       // Init method.
   endwith


  Proc Open
     form.lIsInFocus = false   
       // Don't need to put this in the ReadModal proc, because 
       // modal forms are ALWAYS in focus.
     class::RegisterThisForm()
     form.Init()
     If Class::PreventMultipleConcurrentInstances()
        return
     Endif
     CLASS::RunControlInits(Form)
  return super::open()

  Proc ReadModal
     class::RegisterThisForm()
     form.Init()
     If Class::PreventMultipleConcurrentInstances()
        return
     Endif
     CLASS::RunControlInits(Form)
  return super::ReadModal()

  Proc Init
     /* 
        Over-ride this procedure in forms that are inherited from this 
        class, to do any stuff that needs to happen when the form 
        opens. Use an OnOpen event handler only for the stuff that 
        must be done AFTER a form opens - i.e., stuff that requires 
        a form window handle.
     */

  Proc RunControlInits(InitialObject)
     /* 
        Run the Init methods of any controls that have such methods.
        (This is often faster than running OnOpen events, where the
        form is already open.  This is especially effective for things
        like entryPickers, which build buttons on-the-fly during their
        onOpen/Init methods.)
        For example, your form's Init method could have something 
        like this:
          Form.zEntryField1.Init = CLASS::DoWhatever  // function pointer
          Form.zEntryPicker2.Init = ;
                    {; EntryPicker::Init() ; CLASS::DoWhatever() }
        In some cases, this can give you better performance than 
        using the OnOpen events of the controls. The reason is that 
        when the Init methods are fired the form is not yet open, so 
        there are no GUI controls to refresh.
     */
     private o
     o = InitialObject.First
     do
       if ("CONTAINER"$o.CLASSNAME) OR ("NOTEBOOK"$o.CLASSNAME)
          CLASS::RunControlInits(o)   
             // recurse through the objects of the container
       EndIf
       if Type('o.Init') $ 'FP/CB'   
              // If this control object has an Init method...
          o.Init()
       endif
       o = o.before
     until o.name == InitialObject.First.name
   return


  Proc RegisterThisForm
     /* 
        Create a global reference to this form, for debugging or 
        other purposes:
        (Note: this only holds up for SDI forms.)
     */
     if type('_app.forms') = 'U'
        _app.forms = new object()
       * array()   // an array of form objects, e.g. {{'MyForm',<object>,nCountHowManyCurrentlyOpen},
     endif
     Private cClassName, c
     cClassName = This.ClassName
     c = '_app.forms.' + cClassName + ' = This'
     &c.

  Procedure Repaint   // Call this method to refresh the form.
    /*  
       if argcount()=1
          UpdateWindow(argVector(1))
       else
          UpdateWindow(form.hwnd)   // Or is it better to change the PageNo?
       endif
    */
    local nSavePageNo
    nSavePageNo = Form.PageNo
    Form.PageNo = 99
    Form.PageNo = nSavePageNo   // To solve the refresh problem.
  Return


  Func PreventMultipleConcurrentInstances   
     // Returns True if the form is already open and the user is not 
     // allowed to open it twice concurrently.
     If type('form.AllowMultipleConcurrentInstances')='L' and ;
         not form.AllowMultipleConcurrentInstances
        If type('_app.forms.' + this.ClassName ) = 'O'  
           // If this form has been opened at least once during 
           // this session...
           * Is it open right now?
           If not Empty(form.hWnd)   // If this form has a Windows 
                                     // handle right now...
              msgBox('This form is already open')
              return True
           Endif
        Endif
     Endif
  retu false


  procedure OnGotfocus
     form.lIsInFocus = true
     _app.activeform = form
     on key label f5 _app.activeform.PrintIt()


  procedure OnLostfocus
     form.lIsInFocus = false
     _app.activeform=null
     on key label f5


  procedure PrintIt
     on key label f5     // Prevent multiple printouts for 
                         // users with heavy fingers.
     *keyboard chr(13)   // Uncomment this to choose the default  
                         // printer and avoid the printing dialog.
     _app.activeform.print()
     on key label f5 _app.activeform.Printit()

endclass

/* 
   Note: When creating a new form where zForm is the baseFormClass,
   if you want to resize the form, change its position/size properties 
   at least once in the inspector - otherwise (if you do it just with 
   mouse resizing), it won't "stick".  This is due to a bug in VdB 
   which Borland says will be eliminated in 7.01.
*/

