/*
     GridColoredRow.cc
     This custom class is designed to create a grid that will 
     accept parameters to allow painting row colors based on 
     conditions within the fields of a table. It does this by 
     the use of an array.

     This class requires that an array be created (see example below)
     that must have either three or four columns to it. The 
     columns are:

     Column 1: FieldName -- the name of the field in the table
               Not case sensitive
     Column 2: The Condition (such as "=> 100") -- a standard
               dBASE condition. This will not (at least currently)
               handle such things as trimming a character field.
     Column 3: The color string -- foreground/background, standard
               color options, see HELP for keyword "colorNormal".
     Column 4: Optional -- if the rowSelect or multiSelect properties
               are set to true, another colorstring to change
               the color of the row when it is selected to that
               color, same as above, standard dBASE color strings,
               foreground/background.

     Like anything of this nature, you must set up your conditions
     carefully. If the second condition is such that it overrides
     the first, when you didn't want it to, you may want to reverse
     the condition (and color), for example, as this works in a top
     down fashion, the first condition defined is evaluated -- if true,
     the color is set; if the second condition overrides that, then
     the color changes to the new color ... 

     // Place this code (or something like it) in
     // the form's onOpen event handler OR
     // an overridden Open/ReadModal method:
     // assume the FISH table in the dBASESamples database
     form.GridColoredRow1.OptArray = new Array( 2, 4 ) // 2 conditions, four columns
     // first condition:
     form.GridColoredRow1.OptArray[ 1, 1 ] = "Length CM"  // field
     form.GridColoredRow1.OptArray[ 1, 2 ] = "=> 100"     // condition
     form.GridColoredRow1.OptArray[ 1, 3 ] = "yellow/red" // if condition is true
     form.GridColoredRow1.OptArray[ 1, 4 ] = "red/yellow" // Selected row
     // second condition:
     form.GridColoredRow1.OptArray[ 2, 1 ] = "Length CM"  // field
     form.GridColoredRow1.OptArray[ 2, 2 ] = "=> 200"     // condition
     form.GridColoredRow1.OptArray[ 2, 3 ] = "yellow/green" // if condition is true
     form.GridColoredRow1.OptArray[ 2, 4 ] = "green/yellow" // Selected row
     // more conditions can be added, just make sure to make
     // the array the correct size, etc.

     Caveats:
        The fourth element/column for the SelectedRow color
        is only effective if either the grid's rowSelected
        or multiSelect properties are set to true. If not, 
        then this will have no effect.

*/
class GridColoredRow( oParent ) of Grid( oParent ) custom

   with( this )
      onOpen = class::Grid_onOpen
   endwith

   function Grid_onOpen
      if not form.InDesign // ignore this code if in the designer
         // disallow this to run:
         if type( "this.OptArray" ) == "U" or;
            type( "this.OptArray" ) # "A" // unknown or not an array
            msgbox( "OptArray for grid not assigned, or"+chr(13)+;
                    "is not an array object.",;
                    "Programmer Error!", 16 )
            form.close()
            return
         endif

         // do we have a columns array set?
         if this.columns.size == 0
            msgbox( "Grid column array has not been"+chr(13)+;
                    "set to display specific fields!",;
                    "Programmer Error!", 16 )
            form.close()
            return
         endif

         // we need to assign the event handlers for
         // each grid column's editorControl:
         private i
         for i = 1 to this.columns.size
             this.columns[i].editorControl.beforeCellPaint = class::editorControl_beforeCellPaint
             this.columns[i].editorControl.onCellPaint = class::editorControl_onCellPaint
         next
         this.refresh() // repaint
      endif // form.inDesign
   return

   function editorControl_beforeCellPaint(bSelectedRow)
      /*
         the code here will need to loop through 
         the array, testing each to determine
         color ...
      */
      private oGrid, nCol, oColumn, i, cCmd
      // for initial startup:
      oGrid = this.parent

      // done with that error, we may need to trap each
      // entry in array, too?
      // need to determine # of coumns in array
      nCol = oGrid.OptArray.subscript( oGrid.OptArray.size, 2 )
      oColumn = this.parent

      // only loop through the rows:
      for i = 1 to oGrid.OptArray.size/nCol
          // build link to get to field value, and then
          // compare in the IF statement below:
          cCmd = 'oGrid.dataLink.fields["'+;
                 oGrid.OptArray[i,1]+'"].value '+;
                 oGrid.OptArray[i,2]

          if &cCmd. // if condition is true:

             // if we have < 4 columns, or if we have
             // 4 columns but this is not the currently
             // selected row:
             if ( nCol ==3 ) or ( nCol == 4 and not bSelectedRow )

                // set the color to this:                
                this.colorNormal := oGrid.OptArray[i,3]

             else 
                // if selected and we have a color option,
                // use that for the selected row
                this.colorNormal := oGrid.OptArray[i,4]

             endif // bSelected and all that

          endif // condition is true

      next // next row if there is more than one condition

   return true

   function editorControl_onCellPaint(bSelectedRow)
      // reset color to default for grid:
      this.colorNormal := "" // reset
   return true

endclass