/*

Name: CustomFields.cc

Custom Field classes
By Michael Nuwer
based on suggestions from
Todd Kreuter, Ken Mayer, and Ivar Jessen.

Usage:
function Query_onOpen
    new <type>Field(this.rowset.fields['<FieldName>'])
    return

    Where <FieldName> is the field you want to morph.

dateField(oField)
    Store a date as a string (format is YYYYMMDD); display the
    stored string as a date in a form or report. Useful for
    databases that do not support date fields or support only
    date/time fields.

numberField
    Store a number as a string; display the stored string
    as a number in a form or report.

booleanField(oField)
    Stores a boolean as "Y" or "N"; display the stored string
    as a logical in a form or report. Useful for
    databases that do not support logical fields.

LocgicalToStringField(oField)
    Stores a logical value; display the value as "Yes" or "No"
    Useful for dbf tables where it is desirable to
    display Yes or No strings -- form or reports.

LocgicalToWingdingField(oField)
    Stores a logical value; display the value as chr(80) when
    true and chr(79) when not true (false or null).   
    Useful for dbf tables where it is desirable to
    display logical data with the wingding font's check mark 
    (true) or x (false).

*/

///////////////////////////////////////////////////////
//
// dateField
// Stores a date as a string (format is YYYYMMDD)
// morphs the stored string into a date.
//
///////////////////////////////////////////////////////
Class dateField(oField) of customField(oField) custom

    function beforeGetValue
       local cValue
       cValue = this.parent[this.realField].value
       return dttod(new date(transform(cValue, "@R AAAA AA AA")))

    function canChange( newvalue)
       this.parent[this.realField].value = dtos(newvalue)
       return true

endClass

///////////////////////////////////////////////////////
//
// numberField
// Stores a number as a string
// morphs the stored string into a number.
//
//  Properties:
//     decimals -- sets the number decimal places to
//                 store in the string.
//     pad --      set the char character to pad the
//                 beginning of the string when the
//                 length of the string is less than
//                 the width of the table field.
///////////////////////////////////////////////////////
Class numberField(oField) of customField(oField)

    this.decimals = 0
    this.pad      = ""

    function beforeGetValue
       local cValue
       cValue = this.parent[this.realField].value
       return val(cValue)

    function canChange( newvalue)
       local nLength
       nLength = this.parent[this.realField].length

       this.parent[this.realField].value = ;
          str(newvalue, nLength, this.decimals, this.pad )
       return true

endClass

///////////////////////////////////////////////////////
//
// booleanField
// Stores a boolean as "Y" or "N"
// morphs the stored string into a logical value.
//
///////////////////////////////////////////////////////
Class booleanField(oField) of customField(oField)

    function beforeGetValue
       local cValue
       cValue = this.parent[this.realField].value
       return iif(lower(cValue)="y",true,false)

    function canChange( newvalue)
       this.parent[this.realField].value = ;
            iif(newvalue, "Y", "N")
       return true

endClass

///////////////////////////////////////////////////////
//
// LogicalToStringField
// Stores a boolean
// morphs the stored value to as "Yes" or "No"
//
// Useful for dbf tables where it is desirable to
// display Yes or No strings -- form or reports.
//
///////////////////////////////////////////////////////
Class LogicalToStringField(oField) of customField(oField)

    function beforeGetValue
       local cValue
       bValue = this.parent[this.realField].value
       return iif(bValue, "Yes", "No")

    function canChange( newvalue )
       this.parent[this.realField].value = ;
            iif(lower(newvalue)="Yes", true , false)
       return true

endClass

Class LogicalToWingdingField(oField) of customField(oField)

    function beforeGetValue
       local cValue
       bValue = this.parent[this.realField].value
       return iif(bValue, chr(80), chr(79))

    function canChange( newvalue )
       this.parent[this.realField].value = ;
            iif(newvalue=chr(80), true , false)
       return true

endClass


///////////////////////////////////////////////////////
//
// customField
//
// Base class used by each custom field class.
//
///////////////////////////////////////////////////////
class customField(oField) of Field() custom

    this.realField = oField.fieldName
    this.fieldName = oField.fieldName + "_"

    oField.parent.add(this)


endclass


/*

The following is an example of a user/developer defined
field. The developer must create a class the derives from 
the base CustomField() class and then write the code for the 
beforeGetValue and canChange methods. See CustomFieldsDemo2.wfm
for an example of using this field with radiobuttons.


//   Stored      Display
//     1    <->  Freshman
//     2    <->  Sophomore
//     3    <->  Junior
//     4    <->  Senior

Class MyField(oField) of customField(oField)

    function beforeGetValue
       cValue = this.parent[this.realField].value
       cReturn = ""
       if cValue == "1"
          cReturn = "Freshman"
       elseif cValue == "2"
          cReturn = "Sophomore"
       elseif cValue == "3"
          cReturn = "Junior"
       elseif cValue == "4"
          cReturn = "Senior"
       endif
       return cReturn 

    function canChange( newvalue )
       if newvalue == "Freshman"
          this.parent[this.realField].value = "1"  
       elseif newvalue == "Sophomore"
          this.parent[this.realField].value = "2"  
       elseif newvalue == "Junior"
          this.parent[this.realField].value = "3"  
       elseif newvalue == "Senior"
          this.parent[this.realField].value = "4"  
       endif
       return true

endClass
*/


***** End of custmFields.cc ****


