/*
   Library: dControls.cc
   Author:  Daniel Howard (sproket@total.net)
   History: 
      1.0   December 12, 1998
            Initial release

   Notes:   

      These controls can be implemented anywhere in your hierarchy 
      though they were designed as base controls at the top level.  
      They are designed to give extended behaviours to the base 
      controls.

      They have some visual properties which you can remove if you 
      don't like them.

      These controls or any controls used with the dForm can have 
      an Init method. If one exists it will fire "before" the form 
      opens. Thanks to Peter Rorlick for the idea.

   Custom Properties/Methods: 
      (these belong to all of the objects)

      ParentClassName
         This property is assigned so I can always know what an object
         is no matter how far it gets subclassed.

      FullName()
         Returns the fully qualified name of the object. 
         ie FORM.MYCONTAINER.MYPUSHBUTTON

      SetValue(cPropertyName,xValue)
         This method creates a custom "field" property. This is based
         on Peter Rorlick's and my idea of a way to have setter/getter
         type behaviour for properties. 
         See my ArraySpinBox class for examples of this technique.

      GetValue(cPropertyName)
         This method gets the value from a custom "field" property.
         Same as using this.MyFieldProperty.Value

               
   See each class for a description of specific additions.

*----------------------------------------------------------------*
This software is donated to the public domain. You may modify 
this code in anyway you wish however the author reserves the 
right to whatever official version is published. If you have 
further ideas/bug reports/additions then contact the author and 
your name will be added to the contributors if your idea is 
implemented.

The author accepts no responsibility for the use or consequences 
of the use of this software.
*/


// The standard highlighting I like 
#define DC_CHIGHLIGHT "black/0x73FFFF"
#define DC_BACKGROUND "0x73FFFF"

class DLINE(parentObj) of LINE(parentObj) custom
   this.ParentClassName = "LINE"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DTEXT(parentObj) of TEXT(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "TEXT"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DENTRYFIELD(parentObj) of ENTRYFIELD(parentObj) custom
   with (this)
      colorHighLight = DC_CHIGHLIGHT
      fontSize = 8
   endwith

   this.ParentClassName = "ENTRYFIELD"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue

   function onGotFocus
      this.BorderStyle = 9
   return

   function onLostFocus
      this.borderstyle = 7
   return

endclass

class DPUSHBUTTON(parentObj) of PUSHBUTTON(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "PUSHBUTTON"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DCHECKBOX(parentObj) of CHECKBOX(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "CHECKBOX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass


class DRADIOBUTTON(parentObj) of RADIOBUTTON(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "RADIOBUTTON"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass


class DLISTBOX(parentObj) of LISTBOX(parentObj) custom
   with (this)
      colorHighLight = DC_CHIGHLIGHT
      fontSize = 8
   endwith

   this.ParentClassName = "LISTBOX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DCOMBOBOX(parentObj) of COMBOBOX(parentObj) custom
protect SavedSettings

   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "COMBOBOX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue

	this.SavedSettings = New Object()
      	
   function onGotFocus
      this.SavedSettings.ColorNormal = this.ColorNormal
      this.ColorNormal = DC_CHIGHLIGHT
   return

   function onLostFocus
      if type("this.SavedSettings.ColorNormal") # "U"
         this.ColorNormal = this.SavedSettings.ColorNormal
      endif
   return
endclass

class DTABBOX(parentObj) of TABBOX(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "TABBOX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DRECTANGLE(parentObj) of RECTANGLE(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "RECTANGLE"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DPROGRESS(parentObj) of PROGRESS(parentObj) custom
   this.ParentClassName = "PROGRESS"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DPAINTBOX(parentObj) of PAINTBOX(parentObj) custom
   this.ParentClassName = "PAINTBOX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DNOTEBOOK(parentObj) of NOTEBOOK(parentObj) custom
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "NOTEBOOK"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

/*   
   Class: dTreeView
   Author:  Daniel Howard (sproket@total.net)
   History: 
      1.0   November 23, 1998
            Initial release

   Requires: ini.cc (Bowen Moursund)
             ini2.cc 
             TreeView.WFM 

   Properties:

      PopUpEnable -
                  Replaces the more ambiguous disablePopup property.
                  This is the same as the editor's popup property name
                  so it's a little easier to remember. If set to true
                  (the default) the treeview will call my popup which
                  gives you the extra option to call the TreeView form
                  which allows you to further customize your treeview.
                  I set the disablePopup property to true and I protect
                  it as well because it will conflict with my popup.

   Methods:

      Iterate  -  Loops though all treeitems in the treeview and 
                  executes a code block or function pointer against 
                  each one.

                  Note: This method doesn't work if called in an Init
                  method of a custom treeview. I don't know if this is 
                  a wad or a bug but the firstChild property doesn't 
                  get set until after the treeview has been opened. 
                  The work around is to call it in an OnOpen event of 
                  the treeview or even the Init of the form?

      Fill     -  Takes an array as a parameter and fills the treeview 
                  with the array values. See the TreeDropper sample 
                  form to see how this can work. You can specify 
                  elements of the array as TreeItemInfo objects if you
                  need specific settings for the treeitems to be 
                  applied like image or bold. This method also creates 
                  a "Datasource" property which will be an array of 
                  the same structure that was passed containg pointers
                  to the actual tree items.

                  Note: ALWAYS PASS THE ARRAY BY VALUE!!!
                  a = {"one","two","three"}
                  dTreeView.Fill( (a) )   // Extra BRACKETS!
                  Otherwise you're bound to get unpredictable results!
                  (At least I did!)

      Save     -  Save the status of the treeview to a file. The file
                  name will be the form name + the treeview name + TV.

      Restore. -  Restore the status of a treeview from a file.

   Event Handlers
      OnRightMouseDown -
                  Overrides the default popup as long as PopUpEnable
                  is true (the default).
   Notes:

      There are a few issues with my TreeView. 

      1) If you use the interactive treeview make sure you have no
         treeitems defined in the form designer. There seems to be
         a bug with ReleaseAllChildren() which prevents the treeview
         from completely releasing them. It works ok as long as all
         treeitems are created at run-time. 

      2) The behaviour of my popup is not exacly like the default
         popup. You'll see from the sample form (TestTree.WFM) that
         you'll need to left-click to properly select a treeitem.
         This is an issue with the TreeView itself. RightClicking
         highlights a treeitem but doesn't actually select it.

      3) The TreeItem doesn't always update the Image and 
         SelectedImage properties right away. You can save() and 
         restore() to reset them properly. ImageScaleToFont and
         Indent are also a little wonky. If you can improve on this
         let me know.
      
      4) I found the Save and Restore methodoligy a little ugly. 
         The rumour is that Interbase has implemented this for the 
         next inline so if that's the case than this gets dropped 
         anyway. Also periodically I get a TreeView file not found
         error. I reduced it's occurance by flushing the TV file.
         Wierd. It happens with the TestTree form if you click
         save and restore quickly.
*/
class DTREEVIEW(parentObj) of TREEVIEW(parentObj) custom
protect SaveTreeItems,RestoreTreeItems,IniFile,disablePopUp,GetFileName

   with (this)
      fontSize = 8
      disablePopUp = true
   endwith

   this.PopUpEnable = true
   this.ParentClassName = "TREEVIEW"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
   this.IniFile  = NULL

  
   // Use this custom method of the treeview to iterate through all 
   // items and perform a code block on each.
   Function Iterate
   para cb, o

      if argcount() = 1
         o = this.firstchild
      endif

      private oControl
      oControl = o
      do while oControl # NULL
         cb(oControl)
         if oControl.NoOfChildren > 0
            local oChild
            oChild = oControl.FirstChild
            this.Iterate(cb,oChild)      
         endif      
         oControl = oControl.NextSibling
      enddo
   return 

   // This will take an array as a parameter.
   // Each element can be either a string or another array.
   // This method will fill the treeview according to the array 
   // structure.
   // ReleaseAllChildren() should be called first.

   /* Array Sample
         Test1
         Test2
         { "item1","item2" }
         Test3
         { "item4",{"item5","item6"} }

      Shold look like
      Test1
      Test2
         Item1
         Item2
      Test3
         Item4
            Item5
            Item6                              
   */
   Function Fill
   para a, oParent, aDataSource

      if argcount() = 1 // Top Level
         oParent = this
         this.DataSource = New Array()
         aDataSource = this.DataSource
      endif

      local j, t, cName

      private xItem

      t = oParent
      local n
      n = 1
      for j = 1 to a.Size
         xItem = a[j]
         do case
         case type("xItem") == "A"
            local a2
            a2 = new array(a[j].size)
            acopy(a[j],a2)
            aDataSource.Add( New Array() )
            this.Fill( (a2), t , aDataSource[aDataSource.Size] )
            n = j

         case type("xItem") == "C"
            cName = "TreeItem" + n
            t = New TreeItem(oParent,cName)
            t.Text = a[j]
            aDataSource.Add(t)
            n++
         case type("xItem") == "O"
            cName = "TreeItem" + j
            t = New TreeItem(oParent,cName)
            // This is a TreeItemInfo Object which holds 
            // information about a treeitem. 
            if a[j].ClassName # "TREEITEMINFO"
               t.Text := "Invalid Tree Item Passed!"
            else
               if not empty(a[j].Name)
                  t.Name := a[j].Name
               endif
               t.Bold := a[j].Bold
               t.Checked := a[j].Checked
               t.Image := a[j].Image
               t.SelectedImage := a[j].SelectedImage
               t.Text = a[j].Text
               t.Expanded := a[j].Expanded
               aDataSource.Add(t)               
            endif
            n++
         endcase
      next j
   return


   Function Save
      local cFile

      cFile = this.GetFileName()
      private oldPath
      oldPath = Set("PATH")
      set path to
      if file(cFile)
         erase (cFile)
      endif
      set path to &oldPath.

		set procedure to ini.cc addi
      this.IniFile = New Ini(cFile)

      this.IniFile.SetValue("Treeview","Enabled",this.Enabled+"")
      this.IniFile.SetValue("Treeview","Printable",this.Printable+"")
      this.IniFile.SetValue("Treeview","TabStop",this.TabStop+"")
      this.IniFile.SetValue("Treeview","Visible",this.Visible+"")
      this.IniFile.SetValue("Treeview","FontBold",this.FontBold+"")      
      this.IniFile.SetValue("Treeview","FontItalic",this.FontItalic+"")
      this.IniFile.SetValue("Treeview","FontName","'"+this.FontName+"'")
      this.IniFile.SetValue("Treeview","FontSize",this.FontSize+"")
      this.IniFile.SetValue("Treeview","FontStrikeOut",this.FontStrikeOut+"")
      this.IniFile.SetValue("Treeview","FontUnderLine",this.FontUnderLine+"")
      this.IniFile.SetValue("Treeview","HelpFile","'"+this.HelpFile+"'")
      this.IniFile.SetValue("Treeview","HelpID","'"+this.HelpID+"'")
      this.IniFile.SetValue("Treeview","StatusMessage","'"+this.StatusMessage+"'")
      this.IniFile.SetValue("Treeview","Name","'"+this.Name+"'")
      this.IniFile.SetValue("Treeview","AllowEditLabels",this.AllowEditLabels+"")
      this.IniFile.SetValue("Treeview","AllowEditTree",this.AllowEditTree+"")
      this.IniFile.SetValue("Treeview","CheckBoxes",this.CheckBoxes+"")
      this.IniFile.SetValue("Treeview","CheckedImage","'"+this.CheckedImage+"'")
      this.IniFile.SetValue("Treeview","HasButtons",this.HasButtons+"")
      this.IniFile.SetValue("Treeview","HasLines",this.HasLines+"")
      this.IniFile.SetValue("Treeview","Image","'"+this.Image+"'")
      this.IniFile.SetValue("Treeview","ImageScaleToFont",this.ImageScaleToFont+"")
      this.IniFile.SetValue("Treeview","ImageSize",this.ImageSize+"")
      this.IniFile.SetValue("Treeview","Indent",this.Indent+"")
      this.IniFile.SetValue("Treeview","LinesAtRoot",this.LinesAtRoot+"")
      this.IniFile.SetValue("Treeview","SelectedImage","'"+this.SelectedImage+"'")
      this.IniFile.SetValue("Treeview","ShowSelAlways",this.ShowSelAlways+"")
      this.IniFile.SetValue("Treeview","ToolTips",this.ToolTips+"")
      this.IniFile.SetValue("Treeview","TrackSelect",this.TrackSelect+"")
      this.IniFile.SetValue("Treeview","UnCheckedImage","'"+this.UnCheckedImage+"'")
      this.IniFile.SetValue("Treeview","Anchor",this.Anchor+"")
      this.IniFile.SetValue("Treeview","Height",this.Height+"")
      this.IniFile.SetValue("Treeview","Left",this.Left+"")
      this.IniFile.SetValue("Treeview","Top",this.Top+"")
      this.IniFile.SetValue("Treeview","Width",this.Width+"")
      this.IniFile.SetValue("Treeview","BorderStyle",this.BorderStyle+"")
      this.IniFile.SetValue("Treeview","ColorNormal","'"+this.ColorNormal+"'")
      this.IniFile.SetValue("Treeview","MousePointer",this.MousePointer+"")
      this.IniFile.SetValue("Treeview","PageNo",this.PageNo+"")

      // Custom properties
      this.IniFile.SetValue("Treeview","PopUpEnable",this.PopUpEnable+"")
      this.Iterate(Class::SaveTreeItems)
      this.IniFile.Flush()
   return


   Function SaveTreeItems(o)

      private oParent
      oParent = o.Parent

      // This block gets the full name if the item to save
      // and the full name of its parent.
      local cName,cParent
      cName = o.Name
      cParent = ""
      do while type("oParent.parent") # "U"
         cName = oParent.Name + "." + cName
         cParent = oParent.Name + "." + cParent
         oParent = oParent.Parent
      enddo         
      cName = "FORM."+cName
      cParent = "FORM."+cParent
      cParent = left(cParent,len(cParent)-1) // Chop '.' off
      // Ok Got both names to save

      // Move the parent pointer up to the treeview 
      oParent = o.Parent
      do while oParent.ClassName == "TREEITEM"
         oParent = oParent.Parent
      enddo         

      oParent.IniFile.SetValue(cName,"Name","'"+o.Name+"'")
      oParent.IniFile.SetValue(cName,"Text","'"+o.Text+"'")
      oParent.IniFile.SetValue(cName,"Bold",o.Bold+"")
      oParent.IniFile.SetValue(cName,"Expanded",o.Expanded+"")
      oParent.IniFile.SetValue(cName,"Checked",o.Checked+"")
      oParent.IniFile.SetValue(cName,"Image","'"+o.Image+"'")
      oParent.IniFile.SetValue(cName,"SelectedImage","'"+o.SelectedImage+"'")
   return

   Function Restore()
      local cFile

      cFile = this.GetFileName()
      private oldPath
      oldPath = set("PATH")
      set path to
      if not file(cFile)
         msgbox("TreeView File Not Found!",cFile)
         set path to &oldPath.
         return            
      endif
      set path to &oldPath.

		set proc to ini2.cc addi
      this.IniFile = New Ini2(cFile)

      // The first section is the Treeview itself.
      this.ReleaseAllChildren()

      private m_macro
      local aSection,j
      aSection = this.IniFile.GetSection("Treeview")

      for j = 1 to aSection.Size
         m_macro = this.FullName() + "." + this.IniFile.EntriesArray[j] + " = " + this.IniFile.ValuesArray[j]
         &m_macro.
      next j         
      this.RestoreTreeItems()
      this.IniFile.Flush()
   return

   Function RestoreTreeItems
   local j,k,aSection

      aSection = this.IniFile.GetSections()

      // Start at 2 because the 1st entry is the treeitem itself.
      private m_macro
      private oParent,t
      local n,cName
            
      for j = 2 to aSection.Size
         // Get the Parent
         n = rat(".",aSection[j]) - 1
         m_macro = left(aSection[j],n)
         oParent = &m_macro.

         cName = this.IniFile.GetValue(aSection[j],"Name")

         // Crash and burn if you have treeitems defined in the 
         // form designer!!!!
         t = New TreeItem(oParent,cName)
         this.IniFile.GetSection(aSection[j])
         for k = 2 to this.Inifile.EntriesArray.Size
            m_macro = "t." + this.IniFile.EntriesArray[k] + " := " + this.IniFile.ValuesArray[k]
            &m_macro.
         next k
      next j
      this.Iterate({|o|;o.Expanded := o.Expanded})
   return

   Function GetFileName
      local cFile
      cFile = this.FullName()
		// Remove the dots.....
		local n
      do while true
      	n = at(".",cFile) 
         if n = 0
         	exit
			endif
      	cFile = stuff(cFile,n,1,"")
		enddo
      cFile = form.ClassName + substr(cFile,5,1000) +".TV"
      local cDir
      cDir = set("DIRE")
      if right(cDir,1) # "\"
         cDir += "\"
      endif
      cFile = cDir + cFile
   return ( cFile )

   function onRightMouseDown(flags, col, row)
      if this.PopUpEnable
         if type("form.TreePop") = "U"         
            set procedure to TreeView.POP additive
            form.TreePop = New TreeViewPopUp(form,"TreePop")
            // Custom property on the PopUp which is a refrence to this
            form.TreePop.TreeView = this
         endif
         // Works only with the CHAR metric
         form.TreePop.top = this.Top + row + 1.3
         form.TreePop.left = this.Left + col + 11.3
         if this.firstChild # NULL
            // This seems to have no effect but I'll leave it
            // here in case the bug is fixed.
            form.TreePop.menudelete.Enabled := true
            form.TreePop.menuedit.Enabled := true
         endif
         form.TreePop.Open()
      endif
   return

endclass

class DSLIDER(parentObj) of SLIDER(parentObj) custom
   this.ParentClassName = "SLIDER"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DSCROLLBAR(parentObj) of SCROLLBAR(parentObj) custom
   this.ParentClassName = "SCROLLBAR"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass


class DOLE(parentObj) of OLE(parentObj) custom
   this.ParentClassName = "OLE"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DACTIVEX(parentObj) of ACTIVEX(parentObj) custom
   this.ParentClassName = "ACTIVEX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DREPORTVIEWER(parentObj) of REPORTVIEWER(parentObj) custom
   this.ParentClassName = "REPORTVIEWER"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass


// This class dynamically creates a thin rectangle to cover up the
// unsightly bleeding. Thanks to Richard Sandieson.
class DGRID(parentObj) of GRID(parentObj) custom
   with (this)
      bgColor = DC_BACKGROUND
      borderStyle = 7        // Client
   endwith

   this.ParentClassName = "GRID"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue

   Function Init
      local o
      o = New Rectangle(form)
      o.Top = this.Top
      o.Left = this.Left + 0.1
      o.Width = this.Width - 0.1
      o.Height = 0.2
      this.Rectangle = o
   return

   function onGotFocus
      this.BorderStyle = 1 && 6
   return

   function onLostFocus
      this.BorderStyle = 7
   return

endclass

class DEDITOR(parentObj) of EDITOR(parentObj) custom
protect SavedSettings
   with (this)
      fontSize = 8
   endwith

   this.ParentClassName = "EDITOR"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue

	this.SavedSettings = New Object()

   function onGotFocus
      this.SavedSettings.ColorNormal = this.ColorNormal
      this.ColorNormal = DC_CHIGHLIGHT
   return

   function onLostFocus
      if type("this.SavedSettings.ColorNormal") # "U"
         this.ColorNormal = this.SavedSettings.ColorNormal
      endif
   return
endclass


class DSPINBOX(parentObj) of SPINBOX(parentObj) custom
   with (this)      
      colorHighLight = DC_CHIGHLIGHT
      fontSize = 8
   endwith

   this.ParentClassName = "SPINBOX"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class dSHAPE(parentObj) of SHAPE(parentObj) custom
   this.ParentClassName = "SHAPE"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass

class DCONTAINER(parentObj) of CONTAINER(parentObj) custom
   this.ParentClassName = "CONTAINER"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue

   // Create the setter property
   this.SetValue("Setter",NULL)

   // Hide the border at runtime
   function Init
      this.BorderStyle = 3 // None
   return
endclass

class dIMAGE(parentObj) of IMAGE(parentObj) custom
   this.ParentClassName = "IMAGE"
   this.FullName = DOBJECT::FullName
   this.SetValue = DOBJECT::SetValue
   this.GetValue = DOBJECT::GetValue
endclass


// The good old label object 
CLASS dLabel(FormObj) OF dENTRYFIELD(FormObj) Custom
   this.Border = .F.
   this.TabStop = .F.
   this.ColorNormal = "BtnText/BtnFace"
   this.FontBold = .F.
   this.Height = 1
   this.Width = 26.667
   this.Left = 33.333
   this.Top = 1.1758
   this.mousePointer = 1   // Arrow
   this.Value = "label"
   this.When = {; return false }
ENDCLASS

/*   
   Class: TreeItemInfo
   Author:  Daniel Howard (sproket@total.net)
   History: 
      1.0   November 23, 1998
            Initial release

   This class is simply a container for treeitem
   information. If you pass an array of these objects
   to the fill method of the dTreeView Class it will 
   use this information to fill the treeitems.
*/
Class TreeItemInfo of dObject
   this.Name = ""
   this.Bold = false
   this.Checked = false
   this.Expanded = false
   this.Image = ""
   this.SelectedImage = ""
   this.Text = ""
endclass

/*   
   Class: dObject
   Author:  Daniel Howard (sproket@total.net)
   History: 
      1.0   November 23, 1998
            Initial release

   This is a base object but it has some extra abilities that the
   normal Object doesn't. It's subclassed from DataModRef which means
   that it can be placed in you're component palet. I use this to
   place general logic which can be dropped on the form.
*/
Class dObject of DataModRef
protect active,datamodClass,filename,share

   this.ParentClassName = "DOBJECT"

   // If the custom property RealName is defined use it instead of
   // name.
   Function FullName
      local cName
      if type("this.RealName") # "U"
         cName = this.RealName
      else
         cName = this.Name
      endif

      private o
      o = this
      do while type("o.Parent.Name") # "U"
         cName = o.Parent.Name + "." + cName
      enddo         

      // We're at the form level now 
      cName = "FORM." + cName
   return ( cName )

  	Function GetValue( cPropertyName )   
    	private c
    	c = 'this.'+cPropertyName+'.value'
  	return ( &c. )

   // Recommended alternative syntax to: whatever.value:=something
   // Although both methods will work equally well.
   // 3rd param added for length!  ( not really needed)
  	Function SetValue 
   param cPropertyName, newValue, nLength
   private c
      if type('this.'+cPropertyName) # 'O'
         * Create the new property:
         c = 'this.'+cPropertyName+' = new dField(this)'
         &c.
         local o
         c = 'this.'+cPropertyName
         o = &c.
         o.FieldName := cPropertyName
         if argcount() = 3
            o.Length := nLength
         endif
         c = "this."+cPropertyName
         // Custom property to link this field to the object
*         c = "this."+cPropertyName+".Owner = this"
*         &c.
      endif
      * Now assign the new value:
      c = 'this.'+cPropertyName+'.value'
      &c := newValue
  return
endclass


Class dField(oParent) of Field(oParent)
protect lookupRowset, lookupSQL, readOnly, scale

   this.ParentClassName = "FIELD"   
   if argcount() = 0
      this.Owner = NULL
   else      
      this.Owner = oParent
   endif

/*
   Uncomment the following lines if you want to attach the 
   fields to a phony rowet. The pluses and minuses of this are:

   PLUSES
   * You can datalink your custom field properties to controls.

   MINUSES
   * They will become more strictly typed. The Length propery will
     be respected and they will be automatically padded. (debatable 
     minus)
   * Trying to code with datalinks is much more confusing. It's hard
     to see how/where/why field events are firing.

   I prefer NOT to attach them to a rowset. I find it easier to handle
   datalinks with onChanges instead. That way I have better control
   over the firing of the events.

*/

//    Uncomment these lines
//   if type("_app.PropertyRowset") = "U"
//      _app.PropertyRowset = New PropertyRowset()
//   endif
//   _app.PropertyRowset.Fields.Add(this)
EndClass


Class PropertyRowset(oParent) of Rowset(oParent)
protect autoEdit,endOfSet,filter,filterOptions,indexName, ;
        live,locateOptions,masterfields,masterRowset, ;
        modified, notifyControls, state

protect canAbandon,canAppend,canDelete,canEdit, canGetRow, ;
        canNaigate, canSave, onAbandon, onAppend, onDelete, ;
        onEdit, onNavigate, onSave

protect abandon, applyFilter, applyLocate, atFirst, atLast, ;
        beginAppend, beginEdit, beginFilter, bookmark, ;
        bookmarksEqual, clearFilter, clearRange, count,;
        delete, findKey, findKeyNearest, first, flush, goto, ;
        last, locateNext, lockRow, lockSet, next, refresh, ;
        refreshControls, refreshRow, rowCount, rowNo, save, ;
        setRange, unlock
           
   this.ParentClassName = "ROWSET"
   this.Live := true
   this.State := 2
   this.EndOfSet := false
EndClass



