/*
	---------------------------------------------------------------------------
   	Programmer..: Matthew James BLACK
   	Date........: 7/02/09
   	Version.....: 1.0
   	Notes.......: This is a custom class for a message hashing object using the
                  SHA-224, SHA-256, SHA-384, or SHA-512 algorithm (collectively
                  known as SHA-2).
   	Written for.: dBASEPlus, v2.6.1.5
   	Rev. History: Version 1.0: Initial Creation
   	Calls.......: None
   	Called by...: N/A
   	Usage.......: oRef = new dulux_hash()
    			  cMessageDigest := oRef.hash(<cMessage>[,<nKeySize>])
   	Example.....: cMessageDigest := oRef.hash("abcdefghijklmnop",256)
   	Returns.....: oRef.hash(cMessage) returns cMessageDigest
   	Parameters..: oRef - a Reference to the dulux_hash object
                  cMessage - a Character String containing the text to be
                  	hashed.  The maximum size of cMessage depends upon nKeySize:
                    	nKeySize==224: Maximum len(cMessage)==2^64
                    	nKeySize==256: Maximum len(cMessage)==2^64
                    	nKeySize==384: Maximum len(cMessage)==2^128
                    	nKeySize==512: Maximum len(cMessage)==2^128
                  cMessageDigest - a Character String containing Hexadecimal
					values representing the Message Digest.  cMessageDigest has
                    a size dependant upon nKeySize:
                    	nKeySize==224: len(cMessageDigest)==56
                    	nKeySize==256: len(cMessageDigest)==64
                    	nKeySize==384: len(cMessageDigest)==96
                    	nKeySize==512: len(cMessageDigest)==128
                  nKeySize - the size of the Key in bits.  Must be 224, 256,
                  	384, or 512.  Defaults to 256. 
   --------------------------------------------------------------------------

	Programmer Notes:
	The SHA-2 family of algorithms is one of the hashing algorithms used by
today's computing environment to provide a 1-way hash (or "finger print") of a
message.  Most hashing used today, over the Internet, by Bank ATMs, by Smart
Cards, or whatever, use the SHA-2 family of algorithms or the MD5 algoritm.
Full details of the SHA-2 family of algoritms can be found at
http://csrc.nist.gov/publications/fips/fips180-2/.

	This implementation of the SHA-2 family of algorithms has been written to
provide the same level of security and hashing to dBL applications as can be
found in other languages such as Java and C/C++.  It has been written to use
(by default) the SHA-256 algorithm, but is able to be used with any of the four
SHA-2 algoritms, as desired.  For most applications, the default settings will
be sufficient (cryptographly) for the next several years, and it is recommended
that, unless the application requires greater hashing security, the default
settings be used.

	I encourage others experienced in security and cryptographic techniques to
examine the code for mistakes and to let me know if you find any.  I can be
contacted at:

		matthew@peregrineit.net

	However, changeing code by those not experienced in these areas is highly
discouraged, as, in general, making a small change in hasing code can have
devasting effects on the resulting Message Digest.

	This code is probably not the most efficient implementation of the SHA-2
family of algorithms possible in dBL, but it is as efficient as I can make it.
Again, if you feel you can improve on this (without compromising the
cryptography) please let me know.

	A good hashing algorithm, according to cryptographic experts, is one that
relies only of having the Message kept secret.  In fact, it is advisable that
the actual hashing algorithm be made public knowledge so that other experts can
verify that the hashing algorithm is in fact secure.  The SHA-2 family of
algorithms (and hence this implementation) follows this practice.  In
otherwords, don't be worried about others knowing how your hashing works (ie by
using this Custom Class) as the underlying hashing algorithm has been proved to
be secure (see the SHA-2 family of algorithms documentation for what this
means).

	To fully understand some of the Properties and Methods of the dulux_hash
Class, you need to have an understanding of the SHA-2 family of algorithms.  To
attempt to provide that understanding here would be folly on my part, so I have
instead simply provided the necessary comments to allow an experienced
cryptographer to check my work, and a general dBase developer to use the Class
without having to become an experienced cryptographer.

	(Incedently, I hold a Master's Degree in IT with one of my Majors being
Security.  Although I do not claim to be a cryptographic expert, I have studied
and worked in cryptography, computer forensics, and IT auditing, and so feel
comfortable in producing this Custom Class for use by my fellow dBase
Developers.)

Properties:
	a,b,c,d,e,f,g,h
		Working variables that are the 32-bit or 64-bit words used in the
        computation of the hash value

	cPaddedMessage
		A string, padded with Zeros and the Length of the Original Message in
        Hexadecimal Characters, which contains the Original Message converted to
        Hexadecimal Characters and which has a length a multiple of either 512
        bits or 1024 bits

    h0,h1,h2,h3,h4,h5,h6,h7
		The eight parts of the hash value, each a 32-bit or 64-bit word.

    nKS
		A numeric holding the size of the Key

    k
   		A 1-dimensional array containing the Hashing Constants
	
    w
   		A 1-dimensional array containing the Expanded Message Chunk

Methods:
	addwords32(cWordA,cWordB)
		Used to add two 32-bit Hexidecimal strings together modulo 2^32
		Calls:		None
		Returns:	A 32-bit Hexidecimal string

    addwords64(cWordA,cWordB)
		Used to add two 64-bit Hexidecimal strings together modulo 2^64
		Calls:		addwords32()
		Returns:	A 64-bit Hexidecimal string

    bitand64(cWordA,cWordB)
		Used to bitwise Boolean AND two 64-bit Hexidecimal strings
		Calls:		None
		Returns:	A 64-bit Hexidecimal string

    bitnot64(cWord)
		Used to bitwise Boolean NOT a 64-bit Hexidecimal string
		Calls:		None
		Returns:	A 64-bit Hexidecimal string

    bitxor64(cWordA,cWordB)
		Used to bitwise Boolean Exclusive-OR two 64-bit Hexidecimal strings
		Calls:		None
		Returns:	A 64-bit Hexidecimal string

    bitzrshift64(cWord,nAmount)
		Used to Right Bit-Shift a 64-bit Hexidecimal string by nAmount bits
		Calls:		None
		Returns:	A 64-bit Hexidecimal string

    hash(cMessage,nKeySize)
		Used to hash the Message into the MessageDigest
		Calls:		init_variables()
        			preprocessing()
					hash_chunk256()
					hash_chunk512()
		Returns:	A 56, 64, 96 or 128 character Hexidecimal string

    hash_chunk256(cChunk)
		Used to perform each transformation required for the hashing in each
      	round for the 224-bit and 256-bit hashs
		Calls:		addwords32()
        			rightrotate32()
		Returns:	None

    hash_chunk512(cChunk)
		Used to perform each transformation required for the hashing in each
      	round for the 384-bit and 512-bit hashs
		Calls:		addwords64()
        			bitand64()
                    bitnot64()
                    bitxor64()
                    bitzrshift64()
        			rightrotate64()
		Returns:	None

    init_variables()
   		Used to populate the k array and set the initial values for h0-h7.  The
        values come from the (USA) Federal Information Processing Standards
        Publication 180-2
		Calls:		None
		Returns:	None

    preprocessing(cMessage)
		Used to turn the Message into its Hexidecimal representation, append a
        bit set to "1", append enough bits set to "0" so that the total length
        of the result (cPaddedMessage) is a multiple of 512-bits or 1024-bits
        when the last 32-bit or 64-bit Word is the length of cMessage as a
        Hexidecimal number 
		Calls:		None
		Returns:	None

    rightrotate32(cValue,nAmount)
		Used to bit-wise Right Rotate a 32-bit Hexidecimal string by nAmount
        bits
		Calls:		None
		Returns:	A 32-bit Hexidecimal string

    rightrotate64(cValue,nAmount)
		Used to bit-wise Right Rotate a 64-bit Hexidecimal string by nAmount
        bits
		Calls:		None
		Returns:	A 64-bit Hexidecimal string
   ----------------------------------------------------------------------------
*/

class dulux_hash of object
	// The only unprotected elements are: hash()
   	protect a, b, c, d, e, f, g, h, cPaddedMessage, h0, h1, h2, h3, h4, h5, h6
    protect h7, nKS, k, w
   	protect addwords32, addwords64, bitand64, bitnot64, bitxor64, bitzrshift64
    protect hash_chunk256, hash_chuka512, init_variables, preprocessing
    protect rigtrotate32, rightrotate64

	this.a = space(0)
    this.b = space(0)
    this.c = space(0)
	this.cPaddedMessage = space(0)
    this.d = space(0)
    this.e = space(0)
    this.f = space(0)
    this.g = space(0)
    this.h = space(0)
    this.h0 = space(0)
	this.h1 = space(0)
	this.h2 = space(0)
	this.h3 = space(0)
	this.h4 = space(0)
	this.h5 = space(0)
	this.h6 = space(0)
	this.h7 = space(0)
    this.nKS = 0
    this.k = new array()
    this.w = new array()

    function hash(cMessage,nKeySize)
	/*
	nKeySize is the size of the Key (and thus the size of the resulting Message
    	Digest) in bits.  It must be 224, 256, 384, or 512, which corresponds to
        56, 64, 96, and 128 Hex characters, respectively.  If nKeySize is not
        defined, a value of 256 will be used.
	*/
		local i, nBlocks, cMD
      	i = argcount()
      	if i == 0
			msgbox("No Message Supplied.","Error",48)
			return 1
		elseif i == 1
			this.nKS := 256
		elseif i == 2
			if nKeySize == 224 or nKeySize == 384 or nKeySize == 512
				this.nKS := nKeySize
			else
				this.nKS := 256
			endif
		endif
		class::init_variables()
        class::preprocessing(cMessage)
        if this.nKS == 224 or this.nKS == 256
			this.w.resize(64)
            nBlocks = len(this.cPaddedMessage)/128
			for i=1 to nBlocks
				class::hash_chunk256(substr(this.cPaddedMessage,(i-1)*128+1,128))
        	next
		else
			this.w.resize(80)
			nBlocks = len(this.cPaddedMessage)/256
			for i=1 to nBlocks
				class::hash_chunk512(substr(this.cPaddedMessage,(i-1)*256+1,256))
        	next
        endif
       	cMD = this.h0+this.h1+this.h2+this.h3+this.h4+this.h5
		if this.nKS == 384
        	cMD = cMD
        elseif this.nKS == 224
        	cMD = cMD+this.h6
		else 
        	cMD = cMD+this.h6+this.h7
        endif
	return(cMD)

    function init_variables()
		do case
			case this.nKS == 224
				this.h0 := "c1059ed8"
				this.h1 := "367cd507"
				this.h2 := "3070dd17"
				this.h3 := "f70e5939"
				this.h4 := "ffc00b31"
				this.h5 := "68581511"
				this.h6 := "64f98fa7"
				this.h7 := "befa4fa4"
			case this.nKS == 256
				this.h0 := "6a09e667"
				this.h1 := "bb67ae85"
				this.h2 := "3c6ef372"
				this.h3 := "a54ff53a"
				this.h4 := "510e527f"
				this.h5 := "9b05688c"
				this.h6 := "1f83d9ab"
				this.h7 := "5be0cd19"
			case this.nKS == 384
				this.h0 := "cbbb9d5dc1059ed8"
				this.h1 := "629a292a367cd507"
				this.h2 := "9159015a3070dd17"
				this.h3 := "152fecd8f70e5939"
				this.h4 := "67332667ffc00b31"
				this.h5 := "8eb44a8768581511"
				this.h6 := "db0c2e0d64f98fa7"
				this.h7 := "47b5481dbefa4fa4"
			case this.nKS == 512
				this.h0 := "6a09e667f3bcc908"
				this.h1 := "bb67ae8584caa73b"
				this.h2 := "3c6ef372fe94f82b"
				this.h3 := "a54ff53a5f1d36f1"
				this.h4 := "510e527fade682d1"
				this.h5 := "9b05688c2b3e6c1f"
				this.h6 := "1f83d9abfb41bd6b"
				this.h7 := "5be0cd19137e2179"
		endcase
        if this.nKS == 224 or this.nKS == 256
			this.k.resize(64)
			this.k[1] := "428a2f98" 
			this.k[2] := "71374491" 
			this.k[3] := "b5c0fbcf" 
			this.k[4] := "e9b5dba5" 
			this.k[5] := "3956c25b" 
			this.k[6] := "59f111f1" 
			this.k[7] := "923f82a4" 
			this.k[8] := "ab1c5ed5" 
			this.k[9] := "d807aa98" 
			this.k[10] := "12835b01" 
			this.k[11] := "243185be" 
			this.k[12] := "550c7dc3" 
			this.k[13] := "72be5d74" 
			this.k[14] := "80deb1fe" 
			this.k[15] := "9bdc06a7" 
			this.k[16] := "c19bf174" 
			this.k[17] := "e49b69c1" 
			this.k[18] := "efbe4786" 
			this.k[19] := "0fc19dc6" 
			this.k[20] := "240ca1cc" 
			this.k[21] := "2de92c6f" 
			this.k[22] := "4a7484aa" 
			this.k[23] := "5cb0a9dc" 
			this.k[24] := "76f988da" 
			this.k[25] := "983e5152" 
			this.k[26] := "a831c66d" 
			this.k[27] := "b00327c8" 
			this.k[28] := "bf597fc7" 
			this.k[29] := "c6e00bf3" 
			this.k[30] := "d5a79147" 
			this.k[31] := "06ca6351" 
			this.k[32] := "14292967" 
			this.k[33] := "27b70a85" 
			this.k[34] := "2e1b2138" 
			this.k[35] := "4d2c6dfc" 
			this.k[36] := "53380d13" 
			this.k[37] := "650a7354" 
			this.k[38] := "766a0abb" 
			this.k[39] := "81c2c92e" 
			this.k[40] := "92722c85" 
			this.k[41] := "a2bfe8a1" 
			this.k[42] := "a81a664b" 
			this.k[43] := "c24b8b70" 
			this.k[44] := "c76c51a3" 
			this.k[45] := "d192e819" 
			this.k[46] := "d6990624" 
			this.k[47] := "f40e3585" 
			this.k[48] := "106aa070" 
			this.k[49] := "19a4c116" 
			this.k[50] := "1e376c08" 
			this.k[51] := "2748774c" 
			this.k[52] := "34b0bcb5" 
			this.k[53] := "391c0cb3" 
			this.k[54] := "4ed8aa4a" 
			this.k[55] := "5b9cca4f" 
			this.k[56] := "682e6ff3" 
			this.k[57] := "748f82ee" 
			this.k[58] := "78a5636f" 
			this.k[59] := "84c87814" 
			this.k[60] := "8cc70208" 
			this.k[61] := "90befffa" 
			this.k[62] := "a4506ceb" 
			this.k[63] := "bef9a3f7" 
			this.k[64] := "c67178f2"
        else
			this.k.resize(80)
			this.k[1] := "428a2f98d728ae22" 
			this.k[2] := "7137449123ef65cd" 
			this.k[3] := "b5c0fbcfec4d3b2f" 
			this.k[4] := "e9b5dba58189dbbc" 
			this.k[5] := "3956c25bf348b538" 
			this.k[6] := "59f111f1b605d019" 
			this.k[7] := "923f82a4af194f9b" 
			this.k[8] := "ab1c5ed5da6d8118" 
			this.k[9] := "d807aa98a3030242" 
			this.k[10] := "12835b0145706fbe" 
			this.k[11] := "243185be4ee4b28c" 
			this.k[12] := "550c7dc3d5ffb4e2" 
			this.k[13] := "72be5d74f27b896f" 
			this.k[14] := "80deb1fe3b1696b1" 
			this.k[15] := "9bdc06a725c71235" 
			this.k[16] := "c19bf174cf692694" 
			this.k[17] := "e49b69c19ef14ad2" 
			this.k[18] := "efbe4786384f25e3" 
			this.k[19] := "0fc19dc68b8cd5b5" 
			this.k[20] := "240ca1cc77ac9c65" 
			this.k[21] := "2de92c6f592b0275" 
			this.k[22] := "4a7484aa6ea6e483" 
			this.k[23] := "5cb0a9dcbd41fbd4" 
			this.k[24] := "76f988da831153b5" 
			this.k[25] := "983e5152ee66dfab" 
			this.k[26] := "a831c66d2db43210" 
			this.k[27] := "b00327c898fb213f" 
			this.k[28] := "bf597fc7beef0ee4" 
			this.k[29] := "c6e00bf33da88fc2" 
			this.k[30] := "d5a79147930aa725" 
			this.k[31] := "06ca6351e003826f" 
			this.k[32] := "142929670a0e6e70" 
			this.k[33] := "27b70a8546d22ffc" 
			this.k[34] := "2e1b21385c26c926" 
			this.k[35] := "4d2c6dfc5ac42aed" 
			this.k[36] := "53380d139d95b3df" 
			this.k[37] := "650a73548baf63de" 
			this.k[38] := "766a0abb3c77b2a8" 
			this.k[39] := "81c2c92e47edaee6" 
			this.k[40] := "92722c851482353b" 
			this.k[41] := "a2bfe8a14cf10364" 
			this.k[42] := "a81a664bbc423001" 
			this.k[43] := "c24b8b70d0f89791" 
			this.k[44] := "c76c51a30654be30" 
			this.k[45] := "d192e819d6ef5218" 
			this.k[46] := "d69906245565a910" 
			this.k[47] := "f40e35855771202a" 
			this.k[48] := "106aa07032bbd1b8" 
			this.k[49] := "19a4c116b8d2d0c8" 
			this.k[50] := "1e376c085141ab53" 
			this.k[51] := "2748774cdf8eeb99" 
			this.k[52] := "34b0bcb5e19b48a8" 
			this.k[53] := "391c0cb3c5c95a63" 
			this.k[54] := "4ed8aa4ae3418acb" 
			this.k[55] := "5b9cca4f7763e373" 
			this.k[56] := "682e6ff3d6b2b8a3" 
			this.k[57] := "748f82ee5defb2fc" 
			this.k[58] := "78a5636f43172f60" 
			this.k[59] := "84c87814a1f0ab72" 
			this.k[60] := "8cc702081a6439ec" 
			this.k[61] := "90befffa23631e28" 
			this.k[62] := "a4506cebde82bde9" 
			this.k[63] := "bef9a3f7b2c67915" 
			this.k[64] := "c67178f2e372532b" 
			this.k[65] := "ca273eceea26619c" 
			this.k[66] := "d186b8c721c0c207" 
			this.k[67] := "eada7dd6cde0eb1e" 
			this.k[68] := "f57d4f7fee6ed178" 
			this.k[69] := "06f067aa72176fba" 
			this.k[70] := "0a637dc5a2c898a6" 
			this.k[71] := "113f9804bef90dae" 
			this.k[72] := "1b710b35131c471b" 
			this.k[73] := "28db77f523047d84" 
			this.k[74] := "32caab7b40c72493" 
			this.k[75] := "3c9ebe0a15c9bebc" 
			this.k[76] := "431d67c49c100d4c" 
			this.k[77] := "4cc5d4becb3e42b6" 
			this.k[78] := "597f299cfc657e2a" 
			this.k[79] := "5fcb6fab3ad6faec" 
			this.k[80] := "6c44198c4a475817"
        endif
    return

	function preprocessing(cMessage)
		local i, nMessageLength, nMessageLengthInBits, cMessageLengthInHex, nZeros

        nMessageLength = len(cMessage)
        nMessageLengthInBits = nMessageLength*8
        this.cPaddedMessage = space(0)
		for i=1 to nMessageLength
			this.cPaddedMessage := this.cPaddedMessage+itoh(asc(substr(cMessage,i,1)))
    	next
        if this.nKS == 224 or this.nKS == 256
	        nZeros = (444 - (nMessageLengthInBits%512))/4
			if nZeros < 0
				nZeros = 126 - nZeros
            endif
            cMessageLengthInHex = itoh(nMessageLengthInBits,16)
		else
	        nZeros = (892 - (nMessageLengthInBits%1024))/4
			if nZeros < 0
				nZeros = 254 - nZeros
            endif
            cMessageLengthInHex = itoh(nMessageLengthInBits,32)
        endif
		this.cPaddedMessage := this.cPaddedMessage + "8" + replicate("0",nZeros) + cMessageLengthInHex
    return

	function addwords32(cWordA,cWordB)
	return(itoh((htoi(cWordA)+htoi(cWordB))%4294967296,8))

	function addwords64(cWordA,cWordB)
		local cWordA1, cWordA2, cWordB1, cWordB2, cCarry

        nWordA1 = htoi(substr(cWordA,1,8))
        cWordA2 = substr(cWordA,9,8)
        nWordB1 = htoi(substr(cWordB,1,8))
        cWordB2 = substr(cWordB,9,8)
		nCarry = int((htoi(cWordA2)+htoi(cWordB2))/4294967296)
	return(itoh((nWordA1+nWordB1+nCarry)%4294967296,8)+class::addwords32(cWordA2,cWordB2))

	function bitand64(cWordA,cWordB)
		local nValueA1, nValueA2, nValueB1, nValueB2

        nValueA1 = htoi(substr(cWordA,1,8))
        nValueA2 = htoi(substr(cWordA,9,8))
        nValueB1 = htoi(substr(cWordB,1,8))
        nValueB2 = htoi(substr(cWordB,9,8))
    return(itoh(bitand(nValueA1,nValueB1),8)+itoh(bitand(nValueA2,nValueB2),8))

    function bitnot64(cWord)
		local nValueA, nValueB

        nValueA = htoi(substr(cWord,1,8))
        nValueB = htoi(substr(cWord,9,8))
    return(itoh(bitnot(nValueA),8)+itoh(bitnot(nValueB),8))


	function bitxor64(cWordA,cWordB)
		local nValueA1, nValueA2, nValueB1, nValueB2

        nValueA1 = htoi(substr(cWordA,1,8))
        nValueA2 = htoi(substr(cWordA,9,8))
        nValueB1 = htoi(substr(cWordB,1,8))
        nValueB2 = htoi(substr(cWordB,9,8))
    return(itoh(bitxor(nValueA1,nValueB1),8)+itoh(bitxor(nValueA2,nValueB2),8))

	function bitzrshift64(cWord,nAmount)
		local cValueA, cValueB
    	
        cValueA = htoi(substr(cWord,1,8))
        cValueB = htoi(substr(cWord,9,8))
    return(itoh(bitzrshift(cValueA,nAmount),8)+itoh(bitor(bitzrshift(cValueB,nAmount),bitlshift(cValueA,32-nAmount)),8))

	function rightrotate32(cValue,nAmount)
    return(bitor(bitzrshift(htoi(cValue),nAmount),bitlshift(htoi(cValue),32-nAmount)))

   	function rightrotate64(cValue,nAmount)
		local nValueA, nValueB, cTmp

        nValueA = htoi(substr(cValue,1,8))
        nValueB = htoi(substr(cValue,9,8))
		if nAmount < 32
			cTmp = itoh(bitor(bitzrshift(nValueA,nAmount),bitlshift(nValueB,32-nAmount)),8)+itoh(bitor(bitzrshift(nValueB,nAmount),bitlshift(nValueA,32-nAmount)),8)
        else
			cTmp = itoh(bitor(bitzrshift(nValueB,nAmount-32),bitlshift(nValueA,64-nAmount)),8)+itoh(bitor(bitzrshift(nValueA,nAmount-32),bitlshift(nValueB,64-nAmount)),8)
        endif
    return(cTmp)

    function hash_chunk256(cChunk)
		local ch, i, maj, s0, s1, t1, t2

    	for i=1 to 16
			this.w[i] = substr(cChunk,1+(i-1)*8,8)
	    next
        for i=17 to 64
			s0 = itoh(bitxor(class::rightrotate32(this.w[i-15],7),bitxor(class::rightrotate32(this.w[i-15],18),bitzrshift(htoi(this.w[i-15]),3))),8)
        	s1 = itoh(bitxor(class::rightrotate32(this.w[i-2],17),bitxor(class::rightrotate32(this.w[i-2],19),bitzrshift(htoi(this.w[i-2]),10))),8)
	        this.w[i] = class::addwords32(class::addwords32(this.w[i-16],s0),class::addwords32(this.w[i-7],s1))
		next
		this.a := this.h0
		this.b := this.h1
		this.c := this.h2
		this.d := this.h3
		this.e := this.h4
		this.f := this.h5
		this.g := this.h6
		this.h := this.h7
	    for i=1 to 64
    	    s0 = itoh(bitxor(class::rightrotate32(this.a,2),bitxor(class::rightrotate32(this.a,13),class::rightrotate32(this.a,22))),8)
	        maj = itoh(bitxor(bitand(htoi(this.a),htoi(this.b)),bitxor(bitand(htoi(this.a),htoi(this.c)),bitand(htoi(this.b),htoi(this.c)))),8)
        	t2 = class::addwords32(s0,maj)
    	    s1 = itoh(bitxor(class::rightrotate32(this.e,6),bitxor(class::rightrotate32(this.e,11),class::rightrotate32(this.e,25))),8)
	        ch = itoh(bitxor(bitand(htoi(this.e),htoi(this.f)),bitand(bitnot(htoi(this.e)),htoi(this.g))),8)
        	t1 = class::addwords32(class::addwords32(class::addwords32(this.h,s1),class::addwords32(ch,this.k[i])),this.w[i])
    	    this.h := this.g
	        this.g := this.f
        	this.f := this.e
    	    this.e := class::addwords32(this.d,t1)
	        this.d := this.c
        	this.c := this.b
    	    this.b := this.a
	        this.a := class::addwords32(t1,t2)
		next
		this.h0 := class::addwords32(this.h0,this.a)
		this.h1 := class::addwords32(this.h1,this.b)
		this.h2 := class::addwords32(this.h2,this.c)
		this.h3 := class::addwords32(this.h3,this.d)
		this.h4 := class::addwords32(this.h4,this.e)
		this.h5 := class::addwords32(this.h5,this.f)
	    this.h6 := class::addwords32(this.h6,this.g)
    	this.h7 := class::addwords32(this.h7,this.h)
	return

    function hash_chunk512(cChunk)
		local ch, i, maj, s0, s1, t1, t2

    	for i=1 to 16
			this.w[i] = substr(cChunk,1+(i-1)*16,16)
	    next
        for i=17 to 80
			s0 = class::bitxor64(class::rightrotate64(this.w[i-15],1),class::bitxor64(class::rightrotate64(this.w[i-15],8),class::bitzrshift64(this.w[i-15],7)))
        	s1 = class::bitxor64(class::rightrotate64(this.w[i-2],19),class::bitxor64(class::rightrotate64(this.w[i-2],61),class::bitzrshift64(this.w[i-2],6)))
	        this.w[i] = class::addwords64(class::addwords64(this.w[i-16],s0),class::addwords64(this.w[i-7],s1))
		next
		this.a := this.h0
		this.b := this.h1
		this.c := this.h2
		this.d := this.h3
		this.e := this.h4
		this.f := this.h5
		this.g := this.h6
		this.h := this.h7
	    for i=1 to 80
    	    s0 = class::bitxor64(class::rightrotate64(this.a,28),class::bitxor64(class::rightrotate64(this.a,34),class::rightrotate64(this.a,39)))
	        maj = class::bitxor64(class::bitand64(this.a,this.b),class::bitxor64(class::bitand64(this.a,this.c),class::bitand64(this.b,this.c)))
        	t2 = class::addwords64(s0,maj)
    	    s1 = class::bitxor64(class::rightrotate64(this.e,14),class::bitxor64(class::rightrotate64(this.e,18),class::rightrotate64(this.e,41)))
	        ch = class::bitxor64(class::bitand64(this.e,this.f),class::bitand64(class::bitnot64(this.e),this.g))
        	t1 = class::addwords64(class::addwords64(class::addwords64(this.h,s1),class::addwords64(ch,this.k[i])),this.w[i])
    	    this.h := this.g
	        this.g := this.f
        	this.f := this.e
    	    this.e := class::addwords64(this.d,t1)
	        this.d := this.c
        	this.c := this.b
    	    this.b := this.a
	        this.a := class::addwords64(t1,t2)
		next
		this.h0 := class::addwords64(this.h0,this.a)
		this.h1 := class::addwords64(this.h1,this.b)
		this.h2 := class::addwords64(this.h2,this.c)
		this.h3 := class::addwords64(this.h3,this.d)
		this.h4 := class::addwords64(this.h4,this.e)
		this.h5 := class::addwords64(this.h5,this.f)
	    this.h6 := class::addwords64(this.h6,this.g)
    	this.h7 := class::addwords64(this.h7,this.h)
	return
endclass
