   /*
   --------------------------------------------------------
   Filename.....: fileDlg.cc
   ClassName....: fileDlg
   Purpose......: API Get(Open/Save)FileName dialog
                  with some properties exposed.
   Programmer...: Rick Miller -  http://www.treturn.com/dbase
   Date.........: December 21, 2004
   Notes........: See Notes below.
   Written for..: dBASE dB2K 0.1 and newer.
   Rev. History.: See Revisions below.
   Dependencies.: VStruct32Ex.cc
                  or VStruct.cc,
                  or VStruct.prg,
                  and VStruct.h.
   Usage........: set procedure to fileDlg.cc additive
   HelpFile.....: See OPENFILENAME in win32.hlp file.
   Example:.....: See Examples below.
   SampleFile...: fileDlg.wfm.
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      1)    file property is type "C" when not multiSelect
               or using the putFile method.
      2)    file property is type "A" when multiSelect
               and using the getFile method.
      3)    fileName property holds a fullPath filename of
               the last file selected in the dialog.
      4)    the filter property is made up of 1 or more
               filter strings concantenated together.
      5)    each filter string uses "|" or chr(124).
               between the display name and filetype.
                  example:
                     "All Files (*.*)|*.*"
      6)    each filter string can have multiple file types
               associated with a single display name.
                  example:
                     "dBASE Custom Files|*.cc;*.cfm;*.crp"
            each filetype is separated with a semicolon.
      7)    when using multiple filter strings for the
               filter property separate each filter string
               with "|" or chr(124).
                  example:
                     filterString1 + "|" + filterString2
      8)    the defExt property is added to a typed in
               filename when the typed in filename with and
               without the extension does not already exist.
      9)    property starting with a tilde "~" is protected.
      10)   method starting with "x_" is protected.
   --------------------------------------------------------
   Example 1:  getFile with 1 filetype selection.
   --------------------------------------------------------
      oDlg  =  new fileDlg()
      oDlg.filter := "Custom Class (*.cc)|*.cc"
      cFile =  oDlg.getFile("cc", "Open Custom Class File")
      oDlg.release()
   --------------------------------------------------------
   Example 2:  getFile with multiple filetype selection.
               set the initial filetype selection to cc.
   --------------------------------------------------------
      oDlg  =  new fileDlg()
      oDlg.filter := "Custom Class (*.cc)|*.cc" +;
                     "|Custom Form (*.cfm)|*.cfm"
      cFile =  oDlg.getFile("cc", "Open Custom File")
      oDlg.release()
   --------------------------------------------------------
   Example 3:  getFile with multiple file selection.
   --------------------------------------------------------
      oDlg  =  new fileDlg()
      oDlg.multiSelect  := true
      cFile =  oDlg.getFile()
      if not empty(cFile)
         ? "Selection from: " + oDlg.directory
         for i = 1 oDlg.file.size
            ? "File " + i + ":" + oDlg.file[i]
         endfor
      endif
      oDlg.release()
   --------------------------------------------------------
   Example 4:  putFile with 1 filetype selection.
   --------------------------------------------------------
      oDlg  =  new fileDlg()
      oDlg.filter := "Custom Class (*.cc)|*.cc"
      cFile =  oDlg.putFile("Save Custom Class File", "cc")
      oDlg.release()
   --------------------------------------------------------
   Example 5:  putFile with multiple filetype selection.
   --------------------------------------------------------
      oDlg  =  new fileDlg()
      oDlg.filter := "Custom Class (*.cc)|*.cc"
      oDlg.filter += "|Custom Form (*.cfm)|*.cfm"
      cFile =  oDlg.putFile("Save Custom File", "cc")
      oDlg.release()
   --------------------------------------------------------
   Properties: see OPENFILENAME in win32.hlp file for more.
   --------------------------------------------------------
      createPrompt      <bool>         true
         prompt to create a new file.
      defExt            <char>         ""
         default extension.
      directory         <char>         ""
         default directory.
      dontAddtoRecent   <bool>         false
         do not add to the most recent file list.
      file              <char/array>   ""
         array with multiSelect, char without multiSelect.
      fileMustExist     <bool>         true
         require existence of file.
      fileName          <char>         ""
         fullPath filename of last file.
      filter            <char>         "All Files (*.*)|*.*"
         what appears in the filetype list.
      forceShowHidden   <bool>         false
         show hidden or system files NT 5.??.
      hideReadOnly      <bool>         true
         hide the readOnly checkbox.
      multiSelect       <bool>         false
         allow multiple file selection.
      noChangeDir       <bool>         true
         no directory change allowed.
      noNetworkButton   <bool>         false
         hide the network button.
      openTitle         <char>         "Open File"
         dialog title with getFile().
      overWritePrompt   <bool>         true
         prompt when replacing an existing file.
      pathMustExist     <bool>         true
         require existence of directory.
      saveTitle         <char>         "Save File"
         dialog title with putFile().
   --------------------------------------------------------
   Methods:
   --------------------------------------------------------
      getFile(<char> fileType, <char> title)
         show the file open dialog.
         returns string.
      inspect()
         display object in the inspector.
      putFile(<char> title, <char> fileType)
         show the file save dialog.
         returns string.
      release()
         destroy object. close procedure files.
      reset()
         reset property values to default settings.
   --------------------------------------------------------
   Revisions:
   --------------------------------------------------------
      February 10, 2005
         1) renamed all external functions.
      December 27, 2004
         1) fixed bug in x_ValidateFiletype method.
   */
// include the constants for VStruct.
#include <VStruct.h>

// Flag constants for the openfilename structure.
#define OFN_READONLY                0x00000001
#define OFN_OVERWRITEPROMPT         0x00000002
#define OFN_HIDEREADONLY            0x00000004
#define OFN_NOCHANGEDIR             0x00000008
#define OFN_SHOWHELP                0x00000010
#define OFN_ENABLEHOOK              0x00000020
#define OFN_ENABLETEMPLATE          0x00000040
#define OFN_ENABLETEMPLATEHANDLE    0x00000080
#define OFN_NOVALIDATE              0x00000100
#define OFN_ALLOWMULTISELECT        0x00000200
#define OFN_EXTENSIONDIFFERENT      0x00000400
#define OFN_PATHMUSTEXIST           0x00000800
#define OFN_FILEMUSTEXIST           0x00001000
#define OFN_CREATEPROMPT            0x00002000
#define OFN_SHAREAWARE              0x00004000
#define OFN_NOREADONLYRETURN        0x00008000
#define OFN_NOTESTFILECREATE        0x00010000
#define OFN_NONETWORKBUTTON         0x00020000
#define OFN_NOLONGNAMES             0x00040000  // force no long names for 4.x modules
#define OFN_EXPLORER                0x00080000  // new look commdlg
#define OFN_NODEREFERENCELINKS      0x00100000
#define OFN_LONGNAMES               0x00200000  // force long names for 3.x modules
#define OFN_ENABLEINCLUDENOTIFY     0x00400000  // Win2K and newer.
#define OFN_ENABLESIZING            0x00800000  // Win98 and newer.
#define OFN_DONTADDTORECENT         0x02000000  // Win2K and newer.
#define OFN_FORCESHOWHIDDEN         0x10000000  // Win2K and newer.
#define OFN_SHAREFALLTHROUGH        2
#define OFN_SHARENOWARN             1
#define OFN_SHAREWARN               0

// global alloc.
#define  GMEM_MOVEABLE              0x00000002
#define  GMEM_ZEROINIT              0x00000040

// programmer constants.
#define  FILEDLG_DEFAULTFILTER      "All Files (*.*)|*.*"
#define  FILEDLG_DEFAULTOPENTITLE   "Open File"
#define  FILEDLG_DEFAULTSAVETITLE   "Save File"
#define  FILEDLG_PUBLISHERINFO      "Rick Miller"
#define  FILEDLG_VERSIONINFO        "5.0210"

// programmer macros.
#define  zeros(n)                replicate(chr(0), n)
//---------------------------------------------------------------//
//          constructor.
//---------------------------------------------------------------//
class fileDlg
   this.className =  "fileDlg"
   this.debug     =  false
   this.messages  =  true
   this.versioninfo  =  FILEDLG_VERSIONINFO
   this.publishinfo  =  FILEDLG_PUBLISHERINFO
   this.~runtime  =  iif("run" $ lower(version(0)), true, false)
   this.~procfile =  class::x_GetProcfile()

   if not empty(this.~procFile)
      // protected properties.
      protect  ~memArray, ~procfile, ~runtime

      // protected methods.
      protect  x_CloseProcedure, x_ErrorMessage, x_GetOs,;
               x_GetProcfile, x_InitExterns, x_Initialize,;
               x_InitOpenFileName, x_MakeStringPointer,;
               x_ReleaseFile, x_ReleaseMemory, x_ShowFileDialog,;
               x_ValidateFileFilter, x_ValidateFileSelection,;
               x_ValidateFiletype, x_ValidateFilterIndex

      class::x_InitExterns()
      class::x_Initialize()
      this.osinfo =  class::x_GetOs()
   else
      class::release()
   endif

   //------------------------------------------------------------//
   //       end of constructor   -  methods below.
   //------------------------------------------------------------//
   // cFile    =  char (fullpath filename, filename, or extension).
   // cTitle   =  char (Dialog caption).
   function getFile(cFile, cTitle)
      if type("argVector(1)") == "C"
         if not empty(cFile)
            class::x_ValidateFiletype(cFile)
         endif
      endif
      if type("argVector(2)") == "C"
         if not empty(cTitle)
            this.openTitle := cTitle
         endif
      endif
   return class::x_ShowFileDialog(false)

   //------------------------------------------------------------//
   function inspect
      inspect(this)
   return

   //------------------------------------------------------------//
   // cTitle   =  char (Dialog caption).
   // cFile    =  char (fullpath filename, filename, or extension).
   function putFile(cTitle, cFile)
      if type("argVector(1)") == "C"
         if not empty(cTitle)
            this.saveTitle := cTitle
         endif
      endif
      if type("argVector(2)") == "C"
         if not empty(cFile)
            class::x_ValidateFiletype(cFile)
         endif
      endif
   return class::x_ShowFileDialog(true)

   //------------------------------------------------------------//
   function release
      if type("this.~memArray") == "O"
         release object this.~memArray
         this.~memArray := null
      endif
      class::x_ReleaseFile()
      class::x_CloseProcedure(this.~procfile)
      try
         release object this
      catch(exception e)
      endtry
      class::x_CloseProcedure(program(1))
   return

   //------------------------------------------------------------//
   // reset property values to default settings.
   function reset
      if type("this.~memArray") == "O"
         release object this.~memArray
         this.~memArray := null
      endif
   return class::x_Initialize()

   //------------------------------------------------------------//
   // try and close a procedure file.
   function x_CloseProcedure(cProc)
      try
         close procedure (cProc)
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   function x_ErrorMessage(cBody)
      if lower("" + this.messages) == "true"
         msgbox(cBody, "  fileDlg " + this.version, 64)
      endif
   return

   //------------------------------------------------------------//
   // return some operating system info.
   function x_GetOs
      Local oVers
      oVers =  FILEDLG_Structures::osVersionInfoStructure()
      RMM_GetVersionEx(oVers)
      cRet  =  "" + oVers.getMem('dwMajorVersion')
      cRet  += "." + oVers.getMem('dwMinorVersion')
      cRet  += " Build:" + oVers.getMem('dwBuildNumber')
      cRet  += " " + oVers.getMem('szCSDVersion')
      release object oVers
      oVers := null
   return cRet

   //------------------------------------------------------------//
   function x_GetProcfile
      Local cRet
      cRet  =  ""
      try
         set procedure to VStruct.cc additive
         cRet  := "VStruct.cc"
      catch(exception e)
         try
            set procedure to VStruct32Ex.cc additive
            cRet  := "VStruct32Ex.cc"
         catch(exception e)
            try
               set procedure to VStruct.prg additive
               cRet  := "VStruct.prg"
            catch(exception e)
               try
                  set procedure to :dUFLP:VStruct.prg additive
                  cRet  := ":dUFLP:VStruct.prg"
               catch(exception e)
                  class::x_ErrorMessage("VStruct files are missing.")
               endtry
            endtry
         endtry
      endtry
   return cRet

   //------------------------------------------------------------//
   function x_InitExterns
      if not type("RMM_CopyMemoryFrom") == "FP"
         extern CVOID RMM_CopyMemoryFrom(CPTR, CPTR, CULONG) ;
            kernel32 from "RtlMoveMemory"
      endif
      if not type("RMM_CopyMemoryTo") == "FP"
         extern CVOID RMM_CopyMemoryTo(CLONG, CPTR, CULONG) ;
            kernel32 from "RtlMoveMemory"
      endif
      if not type("RMM_GetActiveWindow") == "FP"
         extern CHANDLE RMM_GetActiveWindow(CVOID) ;
            user32 from "GetActiveWindow"
      endif
      if not type("RMM_GetOpenFilename") == "FP"
         extern CLOGICAL RMM_GetOpenFileName(CPTR) ;
            comdlg32 from "GetOpenFileNameA"
      endif
      if not type("RMM_GetSaveFilename") == "FP"
         extern CLOGICAL RMM_GetSaveFileName(CPTR) ;
            comdlg32 from "GetSaveFileNameA"
      endif
      if not type("RMM_GetVersionEx") == "FP"
         extern CLOGICAL RMM_GetVersionEx(CPTR) ;
            kernel32 from "GetVersionExA"
      endif
      if not type("RMM_GlobalAlloc") == "FP"
         extern CHANDLE RMM_GlobalAlloc(CUINT, CULONG) ;
            kernel32 from "GlobalAlloc"
      endif
      if not type("RMM_GlobalFree") == "FP"
         extern CHANDLE RMM_GlobalFree(CHANDLE) ;
            kernel32 from "GlobalFree"
      endif
      if not type("RMM_GlobalLock") == "FP"
         extern CLONG RMM_GlobalLock(CHANDLE) ;
            kernel32 from "GlobalLock"
      endif
      if not type("RMM_GlobalSize") == "FP"
         extern CULONG RMM_GlobalSize(CHANDLE) ;
            kernel32 from 'GlobalSize'
      endif
      if not type("RMM_GlobalUnLock") == "FP"
         extern CLOGICAL RMM_GlobalUnLock(CHANDLE) ;
            kernel32 from "GlobalUnlock"
      endif
   return

   //------------------------------------------------------------//
   // initialize default property values.
   function x_Initialize
      // custom properties.
      this.directory =  ""
      this.defExt    =  ""
      this.file      =  ""
      this.fileName  =  ""
      this.filter    =  FILEDLG_DEFAULTFILTER
      this.openTitle =  FILEDLG_DEFAULTOPENTITLE
      this.saveTitle =  FILEDLG_DEFAULTSAVETITLE

      // flag properties.
      this.createPrompt    =  true
      this.dontAddtoRecent =  false // for NT 5.0 and up.
      this.fileMustExist   =  true
      this.forceShowHidden =  false // for NT 5.0 and up.
      this.hideReadOnly    =  true
      this.multiSelect     =  false
      this.noChangeDir     =  true
      this.noNetworkButton =  false
      this.overWritePrompt =  true
      this.pathMustExist   =  true

      // holds memory spaces.
      this.~memArray =  new assocArray()
      this.~memArray['lpstrFilter']       =  0
      this.~memArray['lpstrTitle']        =  0
      this.~memArray['lpstrFile']         =  0
      this.~memArray['lpstrInitialDir']   =  0
      this.~memArray['lpstrDefExt']       =  0
   return true

   //------------------------------------------------------------//
   // create openFile structure.
   // create memory pointers.
   // assign pointers to structure.
   // return openFile structure.
   function x_InitOpenFileName(bSave)
      Local cBuff, cFile, nFlag, nSize, oRet, sBuff, sFilt
      // assign variables for use in the structure.
      if bSave
         cBuff =  "" + this.saveTitle
         nSize =  260
      else
         cBuff =  "" + this.openTitle
         nSize =  iif(this.multiSelect, 4096, 260)
      endif
      nFlag =  0
      if this.pathMustExist
         nFlag += OFN_PATHMUSTEXIST
      endif
      if this.fileMustExist
         nFlag += OFN_FILEMUSTEXIST
      endif
      if this.hideReadOnly
         nFlag += OFN_HIDEREADONLY
      endif
      if this.noChangeDir
         nFlag += OFN_NOCHANGEDIR
      endif
      if this.noNetworkButton
         nFlag += OFN_NONETWORKBUTTON
      endif
      if this.osinfo.left(1) == "5"
         if this.dontAddtoRecent
            nFlag += OFN_DONTADDTORECENT
         endif
         if this.forceShowHidden
            nFlag += OFN_FORCESHOWHIDDEN
         endif
      endif
      if bSave
         if this.overWritePrompt
            nFlag += OFN_OVERWRITEPROMPT
         endif
      else
         if this.createPrompt
            nFlag += OFN_CREATEPROMPT
         endif
         if this.multiSelect
            nFlag += OFN_EXPLORER + OFN_ALLOWMULTISELECT
         endif
      endif
      sFilt =  class::x_ValidateFileFilter("" + this.filter)

      // create structure.
      oRet  =  FILEDLG_Structures::openFileNameStructure()

      // assign structure properties.
      oRet.setMem('hwndOwner', RMM_GetActiveWindow())
      // create the filetype filter list.
      oRet.setMem('lpstrFilter',;
         class::x_MakeStringPointer('lpstrFilter',;
         sFilt, sFilt.length + 2))
      // set the filetype filter index.
      oRet.setMem('nFilterIndex',;
         class::x_ValidateFilterIndex(sFilt))
      // set the dialog title.
      oRet.setMem('lpstrTitle', class::x_MakeStringPointer( ;
         'lpstrTitle', cBuff, oRet.getMem('nMaxFileTitle')))
      // set the dialog display options.
      oRet.setMem('Flags', nFlag)
      // set the initially displayed file.
      oRet.setMem('nMaxFile', nSize)
      oRet.setMem('lpstrFile', class::x_MakeStringPointer( ;
         'lpstrFile', "" + iif(type("this.file") == "A",;
         this.file[1], this.file), nSize))
      // set the initially displayed directory.
      oRet.setMem('lpstrInitialDir', class::x_MakeStringPointer( ;
         'lpstrInitialDir', "" + this.directory,;
         this.directory.length + 2))
      // set the default filetype.
      oRet.setMem('lpstrDefExt', class::x_MakeStringPointer( ;
         'lpstrDefExt', "" + this.defExt,;
         this.defExt.length + 2))
   return oRet

   //------------------------------------------------------------//
   // allocate memory(hMem) of size(nLen).
   // create pointer(pMem).
   // copy string data into memory(hMem).
   // return pointer(pMem).
   function x_MakeStringPointer(cKey, cBuff, nLen)
      Local hMem, i, nBuff, pMem, sBuff
      nBuff =  cBuff.length
      sBuff =  zeros(nLen)
      for i = 0 to nLen - 1
         if i < nBuff
            sBuff.setByte(i, asc(cBuff.charAt(i)))
         endif
      endfor
      // allocate memory.
      hMem  =  RMM_GlobalAlloc(GMEM_MOVEABLE + GMEM_ZEROINIT, nLen)
      // get pointer and increment lock count.
      pMem  =  RMM_GlobalLock(hMem)
      // place string data into memory.
      RMM_CopyMemoryTo(pMem, sBuff, nLen)
      // retain a handle to the memory.
      this.~memArray[cKey]  =  hMem
   return int(pMem)

   //------------------------------------------------------------//
   // destroy the file array if it exists.
   function x_ReleaseFile
      if type("this.file") == "A"
         release object this.file
         this.file   := null
      endif
   return

   //------------------------------------------------------------//
   // release memory.
   function x_ReleaseMemory
      Local cKey, hMem
      cKey  =  this.~memArray.firstKey
      for i = 1 to this.~memArray.count()
         hMem  =  this.~memArray[cKey]
         if not hMem == 0
            // decrement the memory lock count.
            RMM_GlobalUnlock(hMem)
            // free the global memory.
            RMM_GlobalFree(hMem)
            this.~memArray[cKey]  =  0
         endif
         cKey  := this.~memArray.nextKey(cKey)
      endfor
   return

   //------------------------------------------------------------//
   // display dialog.
   function x_ShowFileDialog(bSave)
      Local bGet, cDir, cRet, oFile
      cDir  =  set("directory")
      cRet  =  ""
      // get an openfile structure for the function parameter.
      oFile =  class::x_InitOpenFileName(iif(bSave, true, false))
      if bSave
         // show the save dialog.
         bGet  =  RMM_GetSaveFileName(oFile)
      else
         // show the open dialog.
         bGet  =  RMM_GetOpenFileName(oFile)
      endif
      if bGet
         // user selected OK.
         // process the selections.
         Local nSize, pMem, sFile
         pMem  =  int(oFile.getMem('lpstrFile'))
         nSize =  RMM_GlobalSize(this.~memArray['lpstrFile'])
         // create a new string to copy the memory to.
         sFile =  zeros(nSize)
         // copy from memory at pMem to sFile.
         RMM_CopyMemoryFrom(sFile, pMem, nSize)
         // call to parse sFile as needed.
         class::x_ValidateFileSelection(sFile,;
            iif(bSave, false, iif(this.multiSelect, true, false)))
         cRet  := "" + this.fileName
      endif
      // destroy the structure object.
      release object oFile
      oFile := null
      // call to free the memory.
      class::x_ReleaseMemory()
      if this.debug
         class::inspect()
      endif
      if this.noChangeDir
         if not set("directory") == cDir
            set directory to (cDir)
         endif
      endif
   return(cRet)

   //------------------------------------------------------------//
   // replace all "|" with chr(0).
   // return a valid filter string for use with a structure.
   function x_ValidateFileFilter(cFilt)
      if empty(cFilt)
         cFilt := FILEDLG_DEFAULTFILTER
      endif
      Local aFilt, aName, cIndx, cRet, cTmp, i, nPos, nPos1, nPos2
      aFilt =  new array()
      aName =  new array()
      cIndx =  "|"
      cRet  =  ""
      nCnt  =  0
      nLen  =  cFilt.length
      nPos  =  2
      nPos1 =  0
      do while not nPos == -1
         nPos  := cFilt.indexof(cIndx, nPos1)
         if nPos == 0
            // bypass programmer error.
            nPos2 =  nPos
         elseif nPos > nPos1
            // a separator exists.
            nPos2 =  nPos
         elseif nLen > nPos1
            // use the remainder of cFilt
            nPos2 =  nLen
         else
            // bypass string process.
            nPos2 =  nLen - 1
            nPos1 := nLen
         endif
         if nPos2 > nPos1
            // there is a string to process.
            if mod(++ nCnt, 2) == 0
               // add to arrays.
               aName.add(cTmp)
               aFilt.add(cFilt.substring(nPos1, nPos2))
            else
               // start add to arrays.
               cTmp  =  cFilt.substring(nPos1, nPos2)
            endif
         endif
         // reposition the starting search position.
         nPos1 := nPos2 + 1
      enddo
      for i = 1 to aName.size
         if i > 1
            cRet  += chr(0)
         endif
         cRet  += aName[i] + chr(0) + lower(aFilt[i])
      endfor
      release object aFilt
      release object aName
      aFilt := null
      aName := null
   return cRet

   //------------------------------------------------------------//
   function x_ValidateFileSelection(sFile, bMult)
      class::x_ReleaseFile()
      Local cChro, cFile, i, nPos0, nPos1
      cChro =  chr(0)
      cFile =  ""
      // extract sFile to chars.
      for i = 0 to sFile.length - 1
         cFile += chr(sFile.getByte(i))
      endfor
      // find the first chr(0).
      nPos0 =  cFile.indexof(cChro)
      nPos1 =  cFile.indexof(".")
      if bMult
         // create an array.
         this.file   =  new array()
      endif
      if nPos1 > nPos0
         // assign properties for a multiple file selection.
         this.directory := cFile.left(nPos0)
         // the first chr(0) separates the directory
         // from the file list.
         // get the chars following the directory.
         Local cTmp
         cTmp  =  cFile.substring(nPos0 + 1, cFile.length)
         Local nLen, nPos
         nLen  =  cTmp.length
         nPos  =  2
         nPos1 =  0
         // parse cTmp to each filename.
         // each filename is separated with a chr(0).
         do while not nPos == -1
            nPos  := cTmp.indexof(cChro, nPos1)
            if nPos > nPos1
               // add a filename to this.file.
               this.file.add(cTmp.substring(nPos1, nPos))
               nPos1 := nPos + 1
            else
               // exit loop, 1st instance of 2 chr(0) together.
               nPos  := -1
            endif
         enddo
         // the last file selected is in this.file[1].
         this.fileName  := this.directory + "\" + this.file[1]
      else
         this.directory := cFile.left(cFile.lastindexof("\"))
         this.fileName  := cFile.left(nPos0)
         if bMult
            this.file.add( ;
               cFile.substring(this.directory.length + 1, nPos0))
         else
            // assign properties for a single file selection.
            this.file   := ;
               cFile.substring(this.directory.length + 1, nPos0)
         endif
      endif
   return

   //------------------------------------------------------------//
   // evaluate the fileType or fileName passed.
   // update defExt, directory, file, filter where needed
   // to accomodate the passed value.
   function x_ValidateFiletype(cFile)
      Local cExt, nPos1, nPos2
      cExt  =  ""
      nPos1 =  cFile.lastindexof("\")
      nPos2 =  cFile.lastindexof(".")

      if not nPos1 == -1
         // update this.directory.
         this.directory := cFile.left(nPos1)
      endif

      if nPos2 == -1
         // no filename was passed.
         if nPos1 == -1
            if cFile.length < 6
               // use the string.
               cExt  := "" + cFile
            endif
         endif
      else
         // a filename was passed.
         // extract the extension.
         cExt  := cFile.substring(nPos2 + 1, cFile.length)
      endif

      if nPos2 > nPos1 + 1
         // update this.file.
         this.file   := cFile.substring(nPos1 + 1, cFile.length)
      endif

      if not empty(cExt)
         // update this.defExt.
         cExt  := cExt.toLowerCase()
         this.defExt := cExt
         if this.filter.toLowerCase().indexof("|*." + cExt) == -1
            // not in this.filter.
            // add it to this.filter.
            if this.filter.indexof("|") > 0
               // add separator to existing filter.
               this.filter += "|"
            endif
            // add to filter.
            this.filter += ;
               "Custom (*." + cExt + ")|*." + cExt
         endif
      endif
   return

   //------------------------------------------------------------//
   // return the index number of the filter to use in the dialog.
   // default sequence for the index is this.file, this.defExt.
   function x_ValidateFilterIndex(sFilt)
      Local cExt, cFile, nPos, nRet
      // default return value.
      nRet  =  1

      if type("this.file") == "A"
         cFile =  "" + this.file[1]
      else
         cFile =  "" + this.file
      endif
      nPos  =  cFile.lastindexof(".")

      if nPos == -1
         // no extension, use this.defExt.
         cExt  =  "" + this.defExt
      else
         // use this.file extension.
         cExt  =  cFile.substring(nPos + 1, cFile.length)
      endif

      nPos  := sFilt.indexof(chr(0) + "*." + cExt.toLowerCase())
      if not nPos == -1
         // cExt is in the filter.
         Local i, nPos2
         nPos2 =  0
         for i = 1 to 200
            // count the separators.
            nPos2 := sFilt.indexof(chr(0), nPos2 + 1)
            if nPos2 == nPos
               // found it's location.
               // add 1 cuz the first filter has only 1 chr(0).
               nRet  := (i + 1) / 2
               i  := 220
            endif
         endfor
      endif
   return int(nRet)

endclass

//---------------------------------------------------------------//
class FILEDLG_Structures
//---------------------------------------------------------------//
   //------------------------------------------------------------//
   //       end of constructor   -  methods below
   //------------------------------------------------------------//
   function openFileNameStructure
      Local sStruct
      if not VSTRUCT_ISDEF("OPENFILENAME")
         VSTRUCT_DEF('OPENFILENAME',;
         'lStructSize',       TYPE_DWORD,;
         'hwndOwner',         TYPE_CHANDLE,;
         'hInstance',         TYPE_CHANDLE,;
         'lpstrFilter',       TYPE_POINTER,;
         'lpstrCustomFilter', TYPE_POINTER,;
         'nMaxCustFilter',    TYPE_DWORD,;
         'nFilterIndex',      TYPE_DWORD,;
         'lpstrFile',         TYPE_POINTER,;
         'nMaxFile',          TYPE_DWORD,;
         'lpstrFileTitle',    TYPE_POINTER,;
         'nMaxFileTitle',     TYPE_DWORD,;
         'lpstrInitialDir',   TYPE_POINTER,;
         'lpstrTitle',        TYPE_POINTER,;
         'Flags',             TYPE_DWORD,;
         'nFileOffset',       TYPE_WORD,;
         'nFileExtension',    TYPE_WORD,;
         'lpstrDefExt',       TYPE_POINTER,;
         'lCustData',         TYPE_DWORD,;
         'lpfnHook',          TYPE_POINTER,;
         'lpTemplateName',    TYPE_POINTER)
      endif
      sStruct  =  new VStruct('OPENFILENAME')
      // assign default member values.
      // lStructSize should be 76.
      sStruct.setMem('lStructSize', sStruct.getSize())
      sStruct.setMem('nMaxFileTitle', 260)
      /*
      for NT 4.0
         'pvReserved',        TYPE_LPVOID,;  //  *        pvReserved
         'dwReserved',        TYPE_DWORD,;   // DWORD         dwReserved
         'FlagsEx',           TYPE_DWORD)    // DWORD         FlagsEx
      */
   return sStruct

   //------------------------------------------------------------//
   function osVersionInfoStructure
      Local sStruct
      if not VSTRUCT_ISDEF("OSVERSIONINFO")
         VSTRUCT_DEF('OSVERSIONINFO',;
         'dwOSVersionInfoSize',  TYPE_DWORD,;
         'dwMajorVersion',       TYPE_DWORD,;
         'dwMinorVersion',       TYPE_DWORD,;
         'dwBuildNumber',        TYPE_DWORD,;
         'dwPlatformId',         TYPE_DWORD,;
         'szCSDVersion',         TYPE_STRING(128))
      endif
      sStruct  =  new VStruct('OSVERSIONINFO')
      sStruct.setMem('dwOSVersionInfoSize', sStruct.getSize())
   return sStruct

endclass