/* license.cc
   -----------------------------------------------------------------------
   Programmer..: Jan Hoelterling   (jhoelterling@bigfoot.com)
   Date........: June 12, 2000
   Notes.......: A custom class to control the number of users authorized 
                 to use a particular application. Also, get exclusive access
                 and restrict other users from logging in

   Written for.: VDB 7.5
   Rev. History: 06/12/2000 0.1 - Original version

   Calls.......: all encapsulated in this class.
   Called by...: Any

   ---------------------------------------------------------------------

	License counter functionality

	The number of valid users is based on the number of valid records in a 
	standard (level 4) DBF file. A record is valid if it contains a valid 
	license key.

	The class contains the following methods:
	- KeyAdd    - creates a new key in the database
                 creates the license file if it doesn't exist
	- KeyGet    - gets and holds a key (returns true or false for success
	- KeyCheck  - (internal) verify that this key is valid for the serial number
	- KeyGen    - (internal) creates a nnew key

	- Exclusive - locks the application for everybody
	- Unlock    - release the "exclusive" lock

	The structure of the file to use is:
	Structure for table   license_keys.dbf
	Table type            DBASE
	Version               4
	-------------------------------------------------------------------------
	Field  Field Name                        Type          Length  Dec  Index
	    1  LICENSEKEY                        CHARACTER         18           N
	    2  LASTDATE                          DATE               8           N
	    3  LASTTIME                          CHARACTER          5           N
	    4  LASTUSER                          CHARACTER         10           N
	-------------------------------------------------------------------------

	Usage:
	set procedure to license.cc additive

	Step 1 : create the license keys in the file (do only once)
	whatever=new license()
	whatever.KeyAdd() && to create ONE new key, repeat as often as needed

	Step 2 : in the startup part of the app, add the following code
	_app.serial= <any six digit number you want to use>
	_app.user  = user name if you want to track (from a login screen or from

	mlic=new license()
	mlic.serial=_app.serial
	mlic.user = _app.user
	x1=mlic.KeyGet()

	_app.license=mlic && important, otherwise the object is not held open!
	
	if not x1
   	* There were no more licenses left, go bye-bye!
	   quit
	endif

   for exclusive processing, i.e. for a program that PACKs all files:

	x1=msgbox("Do you want to run the day-end processing now?"+chr(13)+;
   	       "All users have to be out of the program",;
      	    "Please confirm",292)
	if x1=6
      x1=_app.license.Exclusive()
	   if x1
         * you now have exclusive access - everybody else will be rejected
         * do your stuff, then don't forget:
         _app.license.Unlock()
         * Now users can log in again
      else
         * you can't get exclusive access 'cause there are still users in
         * and Exclusive() showed you which ones so you can yell at them
      endif
   endif


*/


class license(ParentObj) of object(ParentObj) custom
   this.serial=0             && has to be a 6-digit number
   this.mykey=""             && the license key
   this.user =""             && user's name
   this.license=null
   this.session=null
   this.isexclusive=false    && do you have exclusive access or not?

   function KeyGet
	/*
   -----------------------------------------------------------------------
   Programmer..: Jan Hoelterling
   Date........: June 12, 2000      
   Notes.......: Get the next available license key (if any) and lock the 
                 record

   Written for.: VDB 7.5        
   Rev. History: 06/12/2000 0.1 - Original version 

   Calls.......: None
   Called by...: Any
   Usage.......: obj.KeyGet()
   Example.....: ? obj.KeyGet()
   Returns.....: true or false
   Parameters..: none
   ---------------------------------------------------------------------
	*/

      success=false
      if this.serial=0
         msgbox("Serial number invalid!","Error",16)
         this.mykey=""
      else
         s=new session()
         s.parent           :=this
         this.session       :=s
         s.lockretrycount   := 1
         s.lockretryinterval:= 1

         q= new query()
         this.license       := q 
         q.parent           := this
         q.session          := s
      	q.sql="select * from license_keys.dbf" 
      	q.active=true
         status=0

         do while not q.rowset.endofset
            if q.rowset.lockrow()
               tempkey=q.rowset.fields["licensekey"].value
               if class::keycheck(tempkey)
                  this.mykey=tempkey
                  Status=1
               	q.rowset.fields["lastdate"].value = date()
               	q.rowset.fields["lasttime"].value = time()
               	q.rowset.fields["lastuser"].value = this.user
               	q.rowset.save() && does the lock remain?
                  q.rowset.lockrow()
               	exit
               else
                  q.rowset.unlock()
               endif
            else
               if q.rowset.atfirst()
                  status=-1
                  msgbox("The program is currently locked by '"+;
                         trim(q.rowset.fields["lastuser"].value)+"'"+chr(13)+;
                         "for maintenance purposes. Please try again later.",;
                         "Can't log in at this time",16)
                  exit
               endif
            endif
            q.rowset.next()
         enddo
         if status=0
            msgbox("The maximal number of users is already working"+;
                   " with the program."+chr(13)+;
                   "Please try again later or purchase more licenses.",;
                   "Maximum number of users reached",16)
         else
            if status=1
               success = true
            endif
         endif

      endif
   return(success)
   *-- EoF: KeyGet()

	function KeyAdd
	/*
   -----------------------------------------------------------------------
   Notes.......: Add a license key to the license file

   Returns.....: nothing
   Parameters..: none
   ---------------------------------------------------------------------
	*/
      if this.serial=0
         msgbox("Can't create a license key - Serial Number invalid!",;
                "Error",16)
      else
         if not file("license_keys.dbf")
            create table license_keys (licensekey character(18),;
                                       lastdate   date,;
                                       lasttime   character(5),;
                                       lastuser   character(10) )
            msgbox("License_keys file created","Info")
         endif

  			q=new query()
      	m1=class::KeyGen(this.serial)
      	q.sql="select * from license_keys.dbf" 
      	q.active=true

         mfilter="licensekey = '"+m1+"'"
      	do while q.rowset.applylocate(mfilter)
         	m1=class::keygen(this.serial)
         mfilter="licensekey = '"+m1+"'"
      	enddo
      	q.rowset.beginappend()
      	q.rowset.fields["licensekey"].value = m1
      	q.rowset.save()
      	q.active=false
      	q=null
      endif
   return
   *-- EoF: KeyAdd()


   function KeyCheck
	/*
   -----------------------------------------------------------------------
   Notes.......: Verify that a key in the file is valid, so that the 
                 user can't just add rows to the table to increase his
                 user count
                 This method uses a simple encryption mechanism that is
                 the reverse of KeyGen. Feel free to create your own...

   Returns.....: true or false
   Parameters..: the license key in question
   ---------------------------------------------------------------------
	*/
      parameter mykey
      keygood=true

   	mkey1=right(mykey,5)
   	mkey2=left(mykey,5)
   	msum1=val(subst(mykey,6,4))
   	msum2=val(subst(mykey,10,4))
   	mcserial=right(str(this.serial),6)

   	mcs1=val(left(mcserial,3))
   	mcs2=val(right(mcserial,3))

   	mttl1=msum1
   	for i=1 to 5
      	mttl1=mttl1+asc(subst(mkey1,i))
   	next i
   	if mcs1<>mttl1
      	keygood=false
   	endif

   	mttl2=msum2
   	for i=1 to 5
      	mttl2=mttl2+asc(subst(mkey2,i))
   	next i
   	if mcs2<>mttl2
      	keygood=false
   	endif

	return(keygood)
   *-- EoF: KeyCheck()


	function KeyGen
	/*
   -----------------------------------------------------------------------
   Notes.......: Create a new license key with the serial number given

                 This method uses a simple encryption mechanism. The reverse
                 mechanism (decrypt) is used in KeyCheck. 
                 Feel free to create your own...

   Returns.....: a valid license key
   Parameters..: none
   ---------------------------------------------------------------------
	*/

   	mcserial=right(str(this.serial),6)

   	mcs1=val(left(mcserial,3))
   	mcs2=val(right(mcserial,3))

   	mkey1=""
   	mkey2=""

   	mttl1=0
   	for i=1 to 5
    	   mchar=int(random(0)*90)+32
      	mttl1=mttl1+mchar
      	mkey1=mkey1+chr(mchar)
   	next i
   	msum1=right(str(mcs1-mttl1),4)

   	mttl2=0
   	for i=1 to 5
      	mchar=int(random(0)*90)+32
      	mttl2=mttl2+mchar
      	mkey2=mkey2+chr(mchar)
   	next i
   	msum2=right(str(mcs2-mttl2),4)

   	mkey=mkey2+msum1+msum2+mkey1

	return(mkey)   
   *-- EoF: KeyGen()



   function Exclusive
	/*
   -----------------------------------------------------------------------
   Notes.......: Can you have the tables all to yourself? If not, who is 
                 still in?

   Returns.....: true or false
   Parameters..: none
   ---------------------------------------------------------------------
	*/

      success=false
      if this.serial=0
         msgbox("Serial number invalid!","Error",16)
         this.mykey=""
      else
	      if type("this.license")="U"
  	       	s=new session()
         	s.parent           :=this
         	this.session       :=s
         	s.lockretrycount   := 1
         	s.lockretryinterval:= 1

         	q= new query()
         	this.license       := q 
         	q.parent           := this
         	q.session          := s
      		q.sql="select * from license_keys.dbf" 
      		q.active=true
         else
            q=this.license
      	endif
         q.rowset.first()
         if q.rowset.lockset()
            success = true
            q.rowset.first()
            q.rowset.fields["lastuser"].value = this.user
            q.rowset.save()
            this.isexclusive=true
            msgbox("You now have exclusive access to the application."+chr(13)+;
                   "Everybody trying to log in to this program will receive"+;
                   chr(13)+"a message to this effect.","Exclusive Access",48)
         else
            q.rowset.first()
            if not q.rowset.lockrow()
               status=-1
               msgbox("The program is currently locked by '"+;
                       trim(q.rowset.fields["lastuser"].value)+"'"+chr(13)+;
                      "for maintenance purposes. Please try again later.",;
                      "Can't get exclusive access at this time",16)
            else
               q.rowset.unlock()
	            * find out who's still in
   	         stillin= new array()
     	       	do while not q.rowset.endofset
               	if not q.rowset.lockrow()
                  	stillin.add(q.rowset.fields["lastuser"].value)
               	else
                  	q.rowset.unlock()
               	endif
               	q.rowset.next()
            	enddo
            	mstring="The following users are still logged in: "
            	for i=1 to stillin.size
               	mstring=mstring+chr(13)+"    "+ltrim(str(i))+":   "+;
                          upper(stillin[i])
            	next i
            	msgbox(mstring,"Can't get exclusive access",16)
            endif
         endif
      endif
   return(success)
   *-- EoF: Exclusive()



   function Unlock
	/*
   -----------------------------------------------------------------------
   Notes.......: Unlock the system when you're done with the exclusive
                 stuff.

   Returns.....: true or false
   Parameters..: none
   ---------------------------------------------------------------------
	*/
      success=true
      if not this.isexclusive
         msgbox("Method can only be called if you have exclusive access.",;
                "Error",16)
      else
         q=this.license
         q.rowset.first()
         q.rowset.fields["lastuser"].value=" "
         q.rowset.save()
         q.rowset.unlock()
         this.isexclusive = false
         * now I don't have my license key locked anymore, 
         * so i have to get a new one
         if not class::keyget()
            msgbox("Can't get a license key after exclusive access?","Error",16)
            success=false
         endif
      endif
   return(success)
   *-- EoF: Unlock()

endclass


*-- EOP: license.cc
