CLASS XUPDATESET(sd, dd, RecipientTableName)    // OF UPDATESET custom
/*
   --------------------------------------------------------------------
   Programmer.: Bruce Beacham
   Date.......: 12 August 2000
   Purpose....: Contains a method to append in OODML by name, 
                in the way that append works in XDML.    
                Includes type conversion where possible.
                The values copied onto new rows on the destination 
                (which may already have rows) are those on source 
                fields which have the same name as fields on the 
                destination.    You can specify whether a row is saved 
                or discarded.

   To be completed: Transfer the value of an OLE field.

   Updates:      August 1, 2002 -- Set the xOPFORM as mdi = false 
                 so that its topmost property works.

   Manner of Use:
      o = new xupdateset(datafrom, datato, tablename)
      o.namedAppend(false, true)

      OR
      o = new xupdateset()
      o.source = datafrom
      o.destination = datato
      o.RecipientTableName = tablename
      o.namedAppend(false, true)


   Parameters of the CLASS:
      The source and destination properties must be specified 
      before namedAppend() is called.   They may be specified in the 
      class instantiation, or may be assigned later.
      Either may be a filename or a rowset.

      Parameter RecipientTableName is used in message boxes and 
      is optional.

   Parameters of the principal METHOD namedAppend:
      Condition:  The first parameter of the xappend is an append 
      condition codeblock.   Optional;  in its absence no condition 
      has to be satisfied.   The codeblock should return true or false.
      The method passes the destination rowset as a parameter to 
      the condition codeblock at the stage that the append has been 
      done and the new row is about to be saved;   the codeblock result
      determines whether it is saved or the new row is abandoned.

      Notify:   The second parameter is a logical.   It is whether you 
      wish to be notified by msgbox of the number of rows appended.  
      Optional, default false.   Note that if you use this, you must supply
      a value for the first parameter -- in the example, 'false' is a 
      placeholder for the first parameter.

   Return value of the principal METHOD:
      If successful, the namedAppend method will return the number 
      of rows appended.    Otherwise it returns false.

   Further Notes:
      If the destination given is a rowset, any lookupsql on recipient 
      fields will be respected.

      namedAppend() is the only method call made by the user.
      The other methods are subsidiary and are not called by the user.

   Example:
   Uses the dB2K Fish sample:   Copy this to a .prg and run it.
   * We create a modified structure of the fish table, 
   *  then we append the original fish table's rows onto it.

   truefish = _dbwinhome + "samples\fish.dbf"
   newmodel1 = funique("c:\temp\tmpu????.dbf")
   use (truefish)
   copy to (newmodel1) stru exte
   use (newmodel1) excl
   locate for upper(field_name) = "LENGTH"
   if found()
      replace field_type with "CHARACTER", field_len with 10, ;
         field_dec with 0              // to test type conversion.
   endif
   index on field_name tag fn
   newmodel2 = funique("c:\temp\tmpu????.dbf")
   copy to (newmodel2)     // these fields are in a different order.
   use
   newfish = funique("c:\temp\tmpu????.dbf")
   create (newfish) from (newmodel2)
   close data

   o = new xupdateset(truefish, newfish, "Fish")
   if o.abandon
   else
      done = o.NamedAppend({|rr|(rr.fields["name"].value = "C")}, true)
      // codeblock parameter rr is the destination rowset.
      // done should = 2 rows appended because two in the fish table
      // have names which begin with "C".
      // The method itself will report the number appended
   endif

   use (newfish)
   browse 
   --------------------------------------------------------------------
*/

   protect CheckSource, CheckDestination, xAppendFieldCorrespondence,;
           xAppendThisRowFields

   if not empty(sd)
      this.source = sd
   endif
   if not empty(dd)
      this.destination = dd
   endif
   if empty(RecipientTableName)
      this.RecipientTable = ""
      this.RecipientTablePrefix = ""
   else
      this.RecipientTable = RecipientTableName
      this.RecipientTablePrefix = RecipientTableName + ":  "
   endif

   **** End of constructor code of xUpdateSet *********


   FUNCTION CheckSource
      // Develop a source rowset from what has been given 
      // as osource.
      // Called by the NamedAppend method to ensure it can proceed.

      private src ; src = this.Source
      local abandon ; abandon = false
      local rs, qs
      do case
      case empty(src)
           msgbox("Source not defined", "XUPDATESET:xappend", 16)
           abandon = true
      case type("src") = "O"
           if type("src.classname") # "C"
              msgbox("Source object does not have a classname", ;
                     "XUPDATESET:xappend", 16)
              abandon = true
           endif
           if src.classname # "ROWSET"
              msgbox("Source object is not a Rowset", ;
                     "XUPDATESET:xappend", 16)
              abandon = true
           endif
           rs = this.Source
      case type("src") = "C"
           src = DBFSuffix(src)
           if not file(src)
              msgbox("Source string " + src + " is not a valid filename.", ;
                     "XUPDATESET:xappend", 16)
              abandon = true
           endif
           qs = new query()
           qs.sql = [select * from "] + src + ["]
           qs.active = true
           rs = qs.rowset
      otherwise
           msgbox("Source is neither a rowset nor a filename string.  ";
                  + "Type " + type("src"), "XUPDATESET:xappend", 16)
           abandon = true
      endcase

      this.rs = rs
      this.qs = rs.parent        // the Source's query.

      return (not abandon)


   FUNCTION CheckDestination
      // Develop a destination rowset from what has been given 
      // as o.destination.
      // Called by the NamedAppend method to ensure it can proceed.

      private dest ; dest = this.destination
      local abandon ; abandon = false
      local rd, qd
      do case
      case empty(dest)
           msgbox("Destination not defined", "XUPDATESET:xappend", 16)
           abandon = true
      case type("dest") = "O"
           if type("dest.classname") # "C"
              msgbox("Destination object does not have a classname", ;
                     "XUPDATESET:xappend", 16)
              abandon = true
           endif
           if dest.classname # "ROWSET"
              msgbox("Destination object is not a Rowset", ;
                     "XUPDATESET:xappend", 16)
              abandon = true
           endif
           rd = this.destination
      case type("dest") = "C"
           dest = DBFSuffix(dest)
           if not file(dest)
              msgbox("Destination string " + dest + ;
                     " is not a valid filename.", "XUPDATESET:xappend", 16)
              abandon = true
           endif
           qd = new query()
           qd.sql = [select * from "] + dest + ["]
           qd.active = true
           rd = qd.rowset
      otherwise
           msgbox("Destination is neither a rowset nor a filename string.";
                  + "  Type " + type("dest"), "XUPDATESET:xappend", 16)
           abandon = true
      endcase

      this.rd = rd
      this.qd = rd.parent        // the Destination's query.

      return (not abandon)



   FUNCTION namedAppend(condition, notify, notifyend)
      * Appends like the xdml does, using the field name as 
      *  the correspondent.    So it transcribes any field with 
      *  the same name in source and destination.
      *
      * Condition is a codeblock which will evaluate to an OODML 
      *  logical expression when the destination rowset is passed 
      *  to it as a parameter.
      *	If the parameter is empty (which includes false) 
      *  the condition is not tested.    
      *	If it evaluates false after the append, the row appended 
      *  is not saved.
      *
      * Notify reports the number appended to a message box as
      *  it goes along.   Default false
      *
      * NotifyEnd reports the number appended to a message box
      * and waits for OK.   Default to the value of Notify.
      *
      * Returns the number of rows appended, or false.

      // Check that we have valid source and valid destination rowsets.
      if not (class::CheckSource() and class::CheckDestination())
         return false
      endif

      private pCondition ; pCondition = condition
      local conditionalappend
      conditionalappend = not empty(condition)
      if conditionalappend
         conditionalappend = (type("pCondition") = "CB" or ;
            type("pCondition") = "FP")
         if not conditionalappend
            msgbox("Condition (parameter 1) is not a codeblock or ";
               + "function", "XUPDATESET:xAppend", 16)
            return false
         endif
      endif
      	
      if pcount() < 3
         notifyend = notify
      endif

      * Facilitates further adjustment of the rowsets in customised 
      *  copies of this code.
      local rs, rd
      rs = this.rs
      rd = this.rd

      private rsf, rdf
      rsf = rs.fields
      rdf = rd.fields

      * Get an array a which contains the field names (and types) 
      *	of the fields which appear in both rowsets.
      a = class::xAppendFieldCorrespondence()
      if empty(a[1,1])
         * None correspond!
         return false
      endif

      * Circulate round the source rowset, appending in the 
      *  destination row by row.
      * And within that, circulate round the transferrable fields.
      if not rs.first()
         private rrs ; rrs = rs
         local nm ; nm = ""
         if type("rrs.parent.pathfile") = "C"
            nm = chr(13) + rs.parent.pathfile
         endif
         msgbox(this.RecipientTablePrefix + ;
                "No rows to be appended from the source rowset" + nm,;
                "XUPDATESET:xAppend", 48)
         return false
      endif
      if notify
         local rsform ; rsform = new xOpForm()   // defined below.
         rsform.text = "xAppending"
         rsform.h.visible = true
         rsform.h2.visible = true
         rsform.width += 30
         rsform.h.text = "From:  " + trim(rs.tablename)
         rsform.h2.text = "To:  " + trim(rd.tablename)
         rsform.t.text = trim(this.RecipientTable)
         rsform.open()
         cToPost = ltrim(str(rs.count(), 5))
      endif
      local notifyInterval ; notifyInterval = 10
      a.u = 0
      a.v = 0
      local saverow
      * Each loop of this DO is a fresh row.
      do
         a.v ++
         if not class::xAppendThisRowFields(a)
            return false
         endif
         saverow = true

         if conditionalappend
            saverow = pCondition(rd)
         else
            saverow = true
         endif
         if saverow
            rd.save()
            a.u++
            if notify and mod(a.u, notifyInterval) = 0
               rsform.t.text = "xAppended " + this.RecipientTable ;
                  + ":  " + ltrim(str(a.u, 10)) + " / " + ltrim(str(a.v, 10)) ;
                  + "   of   " + cToPost
            endif
         else
            rd.abandon()
         endif

      until not rs.next()
      if notify
         * In case window persists after closing and interval > 1.
         rsform.t.text = "xAppended " + this.RecipientTable ;
            + ":  " + ltrim(str(a.u, 10)) + " / " + ltrim(str(a.v, 10)) ;
            + "   of   " + cToPost
      endif

      if notify
         rsform.close()
      endif
      if notifyend
         msgbox(this.RecipientTablePrefix + ltrim(str(a.u, 10)) ;
         + " of " + cToPost + " rows appended." ;
         + iif(type("_app.IsDeveloper") = "L" and _app.IsDeveloper, ;
         chr(13)+ "Numbers of fields:  Source = " + rsf.size ;
         + "  Destination = " + rdf.size + "   Correspondence = " ;
         + alen(a, 1), ""), ;
         "xAppend", 64)
      endif

   return (a.u)


   FUNCTION xAppendFieldCorrespondence
      * Creates an array of fields for which values will be transferred.
      * rs is the source rowset, rd is the destination rowset.
      * RecipientTable is an optional identifying string.
      * Make an array of fields, transferrable because they have 
      *  the same name and are of an apt type.  
      * 	And are not calculated fields (viz beforegetvalue).  
      *  But lookupsqls are ok.

      local rs, rd
      rs = this.rs
      rd = this.rd

      private rsf, rdf
      rsf = rs.fields
      rdf = rd.fields

      local a ; a = new array(1, 3)
      * Column 1:  name of field in capitals  2:  source field type   
      *  3:  destination field type
      local s, t, uFieldName
      private fn, rsField, rdField
      t = 0
      for s = 1 to rdf.size
          uFieldName = upper(rdf[s].fieldname)
          fn = 'rsf["' + uFieldName + '"]'
          if type(fn) = "U"
             * How can it be?!
             loop
          endif
          rsField = &fn.
          fn = 'rdf["' + uFieldName + '"]'
          if type(fn) = "U"
             * Destination field doesn't exist on the source.
             loop
          endif
          rdField = &fn.
          if rdField.type = "AUTOINC"
             * Cannot amend this.
             loop
          endif
          if type("rdField.calculated") = "L" and rdField.calculated
             * Identified as such by my routines.
             loop
          endif
          if type("rdField.readOnly") = "L" and rdField.readOnly
             * A dB2K property of a dbfField, set true by my routines.
             loop
          endif
          /* superfluous.      */
          if type("rdField.required") = "U"
             * Probably a calculated field
             loop
          endif*/
          * We have an appendable field.
          if t # 0
             a.grow(1)
          endif
          t++
          a[t, 1] = uFieldName
          a[t, 2] = rsField.type
          a[t, 3] = rdField.type
      endfor

      if t = 0                // can't use array length as it's 2D.
         msgbox(this.RecipientTablePrefix ;
            + "No field names correspond",  ;
            "xAppendFieldCorrespondence", 16)
         a[1,1] = false          // tested by calling program 
                               //for validity of returned array.
      endif

   return (a)


   FUNCTION xAppendThisRowFields(a, NotifyOmitted)
      * Transfer the data in the fields of the source rowset (rs) 
      *  to the destination rowset (rd).
      * a is an array of the corresponding fields.   
      *  Also has a property v, a numeric counter.
      * RecipientTable is an optional identifying string.
      * Does not save the appended rowset - that is for the 
      *  calling program, after tests and any further data transfers.

      local rs, rd
      rs = this.rs
      rd = this.rd
      if not rd.beginAppend()
         * If you get an error in the beginAppend(), ensure you have
         *  no lookupsql on a Long with autonullfields = false.
         * Actually, if not writable a vdb error is generated 
         *  immediately.   Needs catching.
         msgbox(this.RecipientTablePrefix + ;
            "Cannot append to the destination rowset", ;
            "xAppendThisRowFields", 16)
         return false
      endif
      private rsf, rdf, fn, rsField, rdField
      local st, dt
      rsf = rs.fields
      rdf = rd.fields
      for t = 1 to alen(a, 1)      // each t is a field.
          fn = 'rsf["' + a[t, 1] + '"]'
          rsField = &fn.
          fn = 'rdf["' + a[t, 1] + '"]'
          rdField = &fn.

          * Type conversion:
          st = a[t, 2]
          dt = a[t, 3]
          do case
          case st = dt and dt <> "OLE"
               * The easy case! 
               * Includes lookuprowsets and lookupsql-to-lookupsql.
               * Remember - due to a vdb bug it is necessary for 
               *  rd's lookupsql to exclude * fields.
               * If the lookupsql mechanism fails, be sure the destination
               *   rowset's query is using the correct database, the 
               *  same one as the source query uses.
               rdField.value = rsField.value
          case st = "CH" and ;
               (dt = "NU" or dt = "FLOAT" or dt = "LONG")
                rdField.value = val(rsField.value)
          case (st = "NU" or st = "FLOAT" or st = "LONG") ;
               and dt = "CH"
               tidystring = ltrim(trim( str(int(rsField.value), 20, 0) ))
               tidydec = ltrim(trim(str(mod(rsField.value, 1), 12, 10) ))
               tidydec = right(tidydec, len(tidydec) - 1)
               tidystring += tidydec
               do while right(tidystring, 1) = "0"
                  tidystring = left(tidystring, len(tidystring)-1)
               enddo
               if len(ltrim(str(int(rsField.value), 20))) > rdField.length
                  * Not enough room for the integers.
                  rdField.value = repl("*", rdField.length)
               elseif len(tidystring) >= rdField.length
                  rdField.value = tidystring
               else
                  rdField.value = right(space(rdField.length) + tidystring, ;
                     rdField.length)
               endif
          case st = "LOGICAL" and dt = "CH"
               if rdField.length >= 5
                  rdField.value = iif(rsField.value = true, "true", "false")
               elseif rdField.length >= 3
                  rdField.value = iif(rsField.value = true, "Yes", "No")
               else
                  rdField.value = iif(rsField.value = true, "Y", "N")
               endif

          case st = "DATE" and dt = "CH"
               rdField.value = ctod(left(rsField.value, 10))
          case st = "CH" and dt = "DATE"
               rdField.value = dtoc(rsField.value)
          case st = "CH" and dt = "ME"
               rdField.value = rsField.value
          case st = "ME" and dt = "CH"
               rdField.value = rsField.value
          case st = "OL" and dt = "ME"
               rdField.value = rsField.value
          otherwise
               if a.v = 1 and NotifyOmitted      //first row only.
                             // Counted in namedAppend()
                  ? this.RecipientTablePrefix + ;
                    "Not transferred field " + ;
                    a[t, 1] + " type " + a[t, 2]
              endif

         endcase

      endfor           // end of fields loop for this row.
   return true

ENDCLASS


FUNCTION dbfSuffix(t)
   * Appends a .dbf suffix if no suffix exists.
   * Parameter t is an optionally-pathed table name.
   * Used in class xUpdateset.
   * Original is in Library.prg.
   if empty(t)
      msgbox("No table given as a parameter.", "dbfSuffix", 48)
      return false
   endif
   	
   private tt ; tt = ltrim(trim(t))
   local a ; a = new stringex(tt).breakstring("\")
   local b ; b = a[a.size]     // b is the table part of the string.
   local pb ; pb = left(tt, len(tt) - len(b))

   if "." $ b
      * Assume the suffix (or none) is deliberate, so leave alone.
   else
      b = trim(b) + ".dbf"            // nb lower case (for beauty).
   endif
   pb = pb + b

return (pb)


* Used in class xUpdateset to monitor the progress of the append.
CLASS xOPFORM of FORM
   with (this)
      scaleFontBold = false
      scaleFontSize = 8
      height = 5
      width = 50
      text = "Operation in progress..."
      topmost = true
      autoCenter = true
      mdi = false
   endwith


   this.H = new TEXT(this)
   with (this.H)
      height = 1
      left = 0
      top = 0.5
      fontSize = 8
      fontBold = true
      width = (form.width)
      anchor = 2        // Top
      alignVertical = 1	// Middle
      alignHorizontal = 1	// Center
      text = ""
      visible = false
   endwith

   this.H2 = new TEXT(this)
   with (this.H2)
      height = 1
      left = 0
      top = 1.5
      fontSize = 8
      fontBold = true
      width = (form.width)
      anchor = 2        // Top
      alignVertical = 1	// Middle
      alignHorizontal = 1	// Center
      text = ""
      visible = false
   endwith

   this.T = new TEXT(this)
   with (this.T)
      height = 2.5
      left = 0
      top = 3
      fontSize = 8
      width = (form.width)
      anchor = 5           // Centre
      alignVertical = 1	// Middle
      alignHorizontal = 1	// Center
      text = ""
   endwith


ENDCLASS

